import copy
from collections import deque, namedtuple
from typing import Optional, Union, Iterable

import numpy as np

from metadrive.component.map.nuplan_map import NuPlanMap
from metadrive.component.map.scenario_map import ScenarioMap
from metadrive.constants import Decoration, TARGET_VEHICLES
from metadrive.obs.top_down_obs_impl import WorldSurface, VehicleGraphics, LaneGraphics
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.utils.interpolating_line import InterpolatingLine
from metadrive.utils.utils import import_pygame
from metadrive.utils.utils import is_map_related_instance

pygame = import_pygame()

color_white = (255, 255, 255)
history_object = namedtuple("history_object", "name position heading_theta WIDTH LENGTH color done")


def draw_top_down_map(
    map,
    resolution: Iterable = (512, 512),
    draw_drivable_area=True,
    return_surface=False,
    film_size=None,
    reverse_color=False,
    road_color=color_white,
) -> Optional[Union[np.ndarray, pygame.Surface]]:
    import cv2
    film_size = film_size or map.film_size
    surface = WorldSurface(film_size, 0, pygame.Surface(film_size))
    if reverse_color:
        surface.WHITE, surface.BLACK = surface.BLACK, surface.WHITE
        surface.__init__(film_size, 0, pygame.Surface(film_size))
    b_box = map.road_network.get_bounding_box()
    x_len = b_box[1] - b_box[0]
    y_len = b_box[3] - b_box[2]
    max_len = max(x_len, y_len)
    # scaling and center can be easily found by bounding box
    scaling = film_size[1] / max_len - 0.1
    surface.scaling = scaling
    centering_pos = ((b_box[0] + b_box[1]) / 2, (b_box[2] + b_box[3]) / 2)
    surface.move_display_window_to(centering_pos)

    if isinstance(map, ScenarioMap):
        if draw_drivable_area:
            for lane_info in map.road_network.graph.values():
                LaneGraphics.draw_drivable_area(lane_info.lane, surface, color=road_color)
        else:
            for id, data in map.blocks[-1].map_data.items():
                if ScenarioDescription.POLYLINE not in data:
                    continue
                type = data.get("type", None)
                if "boundary" in id:
                    num_seg = int(len(data[ScenarioDescription.POLYLINE]) / 10)
                    for i in range(num_seg):
                        # 10 points
                        end = min((i + 1) * 10, len(data[ScenarioDescription.POLYLINE]))
                        waymo_line = InterpolatingLine(np.asarray(data[ScenarioDescription.POLYLINE][i * 10:end]))
                        LaneGraphics.display_scenario(waymo_line, type, surface)

                    if (i + 1) * 10 < len(data[ScenarioDescription.POLYLINE]):
                        end = len(data[ScenarioDescription.POLYLINE])
                        waymo_line = InterpolatingLine(np.asarray(data[ScenarioDescription.POLYLINE][(i + 1) * 10:end]))
                        LaneGraphics.display_scenario(waymo_line, type, surface)
                else:
                    waymo_line = InterpolatingLine(np.asarray(data[ScenarioDescription.POLYLINE]))
                    LaneGraphics.display_scenario(waymo_line, type, surface)

    elif isinstance(map, NuPlanMap):
        if draw_drivable_area:
            for lane_info in map.road_network.graph.values():
                LaneGraphics.draw_drivable_area(lane_info.lane, surface, color=road_color)
        else:
            for block in map.attached_blocks + [map.boundary_block]:
                for boundary in block.lines.values():
                    line = InterpolatingLine(boundary.points)
                    LaneGraphics.display_nuplan(line, boundary.type, boundary.color, surface)

    else:
        for _from in map.road_network.graph.keys():
            decoration = True if _from == Decoration.start else False
            for _to in map.road_network.graph[_from].keys():
                for l in map.road_network.graph[_from][_to]:
                    if draw_drivable_area:
                        LaneGraphics.draw_drivable_area(l, surface, color=road_color)
                    else:
                        two_side = True if l is map.road_network.graph[_from][_to][-1] or decoration else False
                        LaneGraphics.display(l, surface, two_side, use_line_color=True)
    if return_surface:
        return surface
    ret = cv2.resize(pygame.surfarray.pixels_red(surface), resolution, interpolation=cv2.INTER_LINEAR)
    return ret


def draw_top_down_trajectory(
    surface: WorldSurface, episode_data: dict, entry_differ_color=False, exit_differ_color=False, color_list=None
):
    if entry_differ_color or exit_differ_color:
        assert color_list is not None
    color_map = {}
    if not exit_differ_color and not entry_differ_color:
        color_type = 0
    elif exit_differ_color ^ entry_differ_color:
        color_type = 1
    else:
        color_type = 2

    if entry_differ_color:
        # init only once
        if "spawn_roads" in episode_data:
            spawn_roads = episode_data["spawn_roads"]
        else:
            spawn_roads = set()
            first_frame = episode_data["frame"][0]
            for state in first_frame[TARGET_VEHICLES].values():
                spawn_roads.add((state["spawn_road"][0], state["spawn_road"][1]))
        keys = [road[0] for road in list(spawn_roads)]
        keys.sort()
        color_map = {key: color for key, color in zip(keys, color_list)}

    for frame in episode_data["frame"]:
        for k, state, in frame[TARGET_VEHICLES].items():
            if color_type == 0:
                color = color_white
            elif color_type == 1:
                if exit_differ_color:
                    key = state["destination"][1]
                    if key not in color_map:
                        color_map[key] = color_list.pop()
                    color = color_map[key]
                else:
                    color = color_map[state["spawn_road"][0]]
            else:
                key_1 = state["spawn_road"][0]
                key_2 = state["destination"][1]
                if key_1 not in color_map:
                    color_map[key_1] = dict()
                if key_2 not in color_map[key_1]:
                    color_map[key_1][key_2] = color_list.pop()
                color = color_map[key_1][key_2]
            start = state["position"]
            pygame.draw.circle(surface, color, surface.pos2pix(start[0], start[1]), 1)
    for step, frame in enumerate(episode_data["frame"]):
        for k, state in frame[TARGET_VEHICLES].items():
            if not state["done"]:
                continue
            start = state["position"]
            if state["done"]:
                pygame.draw.circle(surface, (0, 0, 0), surface.pos2pix(start[0], start[1]), 5)
    return surface


class TopDownRenderer:
    def __init__(
        self,
        film_size=(1000, 1000),
        screen_size=(1000, 1000),
        light_background=True,
        num_stack=15,
        history_smooth=0,
        road_color=(80, 80, 80),
        show_agent_name=False,
        camera_position=None,
        target_vehicle_heading_up=False,
        draw_target_vehicle_trajectory=False,
        **kwargs
        # current_track_vehicle=None
    ):
        # Setup some useful flags
        self.position = camera_position
        self.target_vehicle_heading_up = target_vehicle_heading_up
        self.show_agent_name = show_agent_name
        self.draw_target_vehicle_trajectory = draw_target_vehicle_trajectory

        if self.show_agent_name:
            pygame.init()

        # self.engine = get_engine()
        # self._screen_size = screen_size
        self.pygame_font = None
        self.map = self.engine.current_map
        self.stack_frames = deque(maxlen=num_stack)
        self.history_objects = deque(maxlen=num_stack)
        self.history_target_vehicle = []
        self.history_smooth = history_smooth
        # self.current_track_vehicle = current_track_vehicle
        if self.target_vehicle_heading_up:
            assert self.current_track_vehicle is not None, "Specify which vehicle to track"
        self.road_color = road_color
        self._light_background = light_background
        self._text_render_pos = [50, 50]
        self._font_size = 25
        self._text_render_interval = 20

        # Setup the canvas
        # (1) background is the underlying layer. It is fixed and will never change unless the map changes.
        self._background_canvas = draw_top_down_map(
            self.map,
            draw_drivable_area=False,
            return_surface=True,
            film_size=film_size,
            road_color=road_color,
        )
        if self._light_background:
            pixels = pygame.surfarray.pixels2d(self._background_canvas)
            pixels ^= 2**32 - 1
            del pixels
        # (2) runtime is a copy of the background so you can draw movable things on it. It is super large
        # and our vehicles can draw on this large canvas.
        self._runtime_canvas = self._background_canvas.copy()

        # (3) it is only used for track vehicle
        self.receptive_field_double = (
            int(self._runtime_canvas.pix(100 * np.sqrt(2))) * 2, int(self._runtime_canvas.pix(100 * np.sqrt(2))) * 2
        )
        self.canvas_rotate = pygame.Surface(self.receptive_field_double)
        # self._runtime_output = self._background_canvas.copy()  # TODO(pzh) what is this?

        # Setup some runtime variables
        self._render_size = screen_size
        self._background_size = tuple(self._background_canvas.get_size())
        # screen_size = self._screen_size or self._render_size
        # self._blit_size = (int(screen_size[0] * self._zoomin), int(screen_size[1] * self._zoomin))
        # self._blit_rect = (
        #     -(self._blit_size[0] - screen_size[0]) / 2, -(self._blit_size[1] - screen_size[1]) / 2, screen_size[0],
        #     screen_size[1]
        # )

        # screen and canvas are a regional surface where only part of the super large background will draw.
        # (3) screen is the popup window and canvas is a wrapper to screen but with more features
        self._render_canvas = pygame.display.set_mode(self._render_size)
        self._render_canvas.set_alpha(None)
        self._render_canvas.fill(color_white)

        # self.canvas = self._render_canvas
        # self.canvas = pygame.Surface(self._render_canvas.get_size())

        # Draw
        self.blit()
        self.kwargs = kwargs

        # key accept
        self.need_reset = False

    @property
    def canvas(self):
        return self._render_canvas

    def refresh(self):
        self._runtime_canvas.blit(self._background_canvas, (0, 0))
        self.canvas.fill(color_white)

    def render(self, text, *args, **kwargs):
        self.need_reset = False
        key_press = pygame.key.get_pressed()
        if key_press[pygame.K_r]:
            self.need_reset = True

        # Record current target vehicle
        objects = self.engine.get_objects(lambda obj: not is_map_related_instance(obj))
        this_frame_objects = self._append_frame_objects(objects)
        self.history_objects.append(this_frame_objects)

        if self.draw_target_vehicle_trajectory:
            self.history_target_vehicle.append(
                history_object(
                    name=self.current_track_vehicle.name,
                    heading_theta=self.current_track_vehicle.heading_theta,
                    WIDTH=self.current_track_vehicle.top_down_width,
                    LENGTH=self.current_track_vehicle.top_down_length,
                    position=self.current_track_vehicle.position,
                    color=self.current_track_vehicle.top_down_color,
                    done=False
                )
            )

        self._handle_event()
        self.refresh()
        self._draw(*args, **kwargs)
        self._add_text(text)
        self.blit()
        ret = self.canvas.copy()
        ret = ret.convert(24)
        return ret

    def _add_text(self, text: dict):
        if not text:
            return
        if not pygame.get_init():
            pygame.init()
        font2 = pygame.font.SysFont('didot.ttc', 25)
        # pygame do not support multiline text render
        count = 0
        for key, value in text.items():
            line = str(key) + ":" + str(value)
            img2 = font2.render(line, True, (0, 0, 0))
            this_line_pos = copy.copy(self._text_render_pos)
            this_line_pos[-1] += count * self._text_render_interval
            self._render_canvas.blit(img2, this_line_pos)
            count += 1

    def blit(self):
        # self._render_canvas.blit(self._runtime_canvas, (0, 0))
        pygame.display.update()

    def close(self):
        pygame.quit()

    def reset(self, map):
        # Reset the super large background
        self._background_canvas = draw_top_down_map(
            map,
            draw_drivable_area=False,
            return_surface=True,
            film_size=self._background_size,
            road_color=self.road_color,
        )
        self._light_background = self._light_background
        if self._light_background:
            pixels = pygame.surfarray.pixels2d(self._background_canvas)
            pixels ^= 2**32 - 1
            del pixels

        # Reset several useful variables.
        # self._render_size = self._background_canvas.get_size()
        # Maybe we can optimize here! We don't need to copy but just blit new background on it.

        self._runtime_canvas = self._background_canvas.copy()
        self.canvas_rotate = pygame.Surface(self.receptive_field_double)

        # self._runtime_output = self._background_canvas.copy()
        self._background_size = tuple(self._background_canvas.get_size())

        self.history_objects.clear()
        self.history_target_vehicle.clear()

    @property
    def current_track_vehicle(self):
        return self.engine.current_track_vehicle

    def _append_frame_objects(self, objects):
        frame_objects = []
        for name, obj in objects.items():
            frame_objects.append(
                history_object(
                    name=name,
                    heading_theta=obj.heading_theta,
                    WIDTH=obj.top_down_width,
                    LENGTH=obj.top_down_length,
                    position=obj.position,
                    color=obj.top_down_color,
                    done=False
                )
            )
        return frame_objects

    def _draw(self, *args, **kwargs):
        """
        This is the core function to process the
        """
        if len(self.history_objects) == 0:
            return

        for i, objects in enumerate(self.history_objects):
            i = len(self.history_objects) - i
            if self.history_smooth != 0 and (i % self.history_smooth != 0):
                continue
            for v in objects:
                c = v.color
                h = v.heading_theta
                h = h if abs(h) > 2 * np.pi / 180 else 0
                x = abs(int(i))
                alpha_f = x / len(self.history_objects)
                VehicleGraphics.display(
                    vehicle=v,
                    surface=self._runtime_canvas,
                    heading=h,
                    color=(c[0] + alpha_f * (255 - c[0]), c[1] + alpha_f * (255 - c[1]), c[2] + alpha_f * (255 - c[2])),
                    draw_countour=False
                )

        # Draw the whole trajectory of ego vehicle with no gradient colors:
        if self.draw_target_vehicle_trajectory:
            for i, v in enumerate(self.history_target_vehicle):
                i = len(self.history_target_vehicle) - i
                if self.history_smooth != 0 and (i % self.history_smooth != 0):
                    continue
                c = v.color
                h = v.heading_theta
                h = h if abs(h) > 2 * np.pi / 180 else 0
                x = abs(int(i))
                alpha_f = min(x / len(self.history_target_vehicle), 0.5)
                # alpha_f = 0
                VehicleGraphics.display(
                    vehicle=v,
                    surface=self._runtime_canvas,
                    heading=h,
                    color=(c[0] + alpha_f * (255 - c[0]), c[1] + alpha_f * (255 - c[1]), c[2] + alpha_f * (255 - c[2])),
                    draw_countour=False
                )

        # Draw current vehicle with black contour
        # Use this line if you wish to draw "future" trajectory.
        # i is the index of vehicle that we will render a black box for it.
        # i = int(len(self.history_vehicles) / 2)
        i = -1
        for v in self.history_objects[i]:
            h = v.heading_theta
            c = v.color
            h = h if abs(h) > 2 * np.pi / 180 else 0
            alpha_f = 0
            VehicleGraphics.display(
                vehicle=v,
                surface=self._runtime_canvas,
                heading=h,
                color=(c[0] + alpha_f * (255 - c[0]), c[1] + alpha_f * (255 - c[1]), c[2] + alpha_f * (255 - c[2])),
                draw_countour=True,
                contour_width=2
            )

        if not hasattr(self, "_deads"):
            self._deads = []

        for v in self._deads:
            pygame.draw.circle(
                surface=self._runtime_canvas,
                color=(255, 0, 0),
                center=self._runtime_canvas.pos2pix(v.position[0], v.position[1]),
                radius=5
            )

        for v in self.history_objects[i]:
            if v.done:
                pygame.draw.circle(
                    surface=self._runtime_canvas,
                    color=(255, 0, 0),
                    center=self._runtime_canvas.pos2pix(v.position[0], v.position[1]),
                    radius=5
                )
                self._deads.append(v)

        v = self.current_track_vehicle
        canvas = self._runtime_canvas
        field = self._render_canvas.get_size()
        if not self.target_vehicle_heading_up:
            if self.position is not None or v is not None:
                cam_pos = (self.position or v.position)
                position = self._runtime_canvas.pos2pix(*cam_pos)
            else:
                position = (field[0] / 2, field[1] / 2)
            off = (position[0] - field[0] / 2, position[1] - field[1] / 2)
            self.canvas.blit(source=canvas, dest=(0, 0), area=(off[0], off[1], field[0], field[1]))
        else:
            position = self._runtime_canvas.pos2pix(*v.position)
            self.canvas_rotate.blit(
                canvas, (0, 0), (
                    position[0] - self.receptive_field_double[0] / 2, position[1] - self.receptive_field_double[1] / 2,
                    self.receptive_field_double[0], self.receptive_field_double[1]
                )
            )

            rotation = np.rad2deg(v.heading_theta) + 90
            new_canvas = pygame.transform.rotozoom(self.canvas_rotate, rotation, 1)

            size = self._render_canvas.get_size()
            self._render_canvas.blit(
                new_canvas,
                (0, 0),
                (
                    new_canvas.get_size()[0] / 2 - size[0] / 2,  # Left
                    new_canvas.get_size()[1] / 2 - size[1] / 2,  # Top
                    size[0],  # Width
                    size[1]  # Height
                )
            )

        if "traffic_light_msg" in kwargs:
            raise ValueError("This function is broken")
            if kwargs["traffic_light_msg"] < 0.5:
                traffic_light_color = (0, 255, 0)
            else:
                traffic_light_color = (255, 0, 0)
            pygame.draw.circle(
                surface=self.canvas,
                color=traffic_light_color,
                center=(self.canvas.get_size()[0] * 0.1, self.canvas.get_size()[1] * 0.1),
                radius=20
            )

        if self.show_agent_name:
            raise ValueError("This function is broken")
            # FIXME check this later
            if self.pygame_font is None:
                self.pygame_font = pygame.font.SysFont("Arial.ttf", 30)
            agents = [agent.name for agent in list(self.engine.agents.values())]
            for v in self.history_objects[i]:
                if v.name in agents:
                    position = self._runtime_canvas.pos2pix(*v.position)
                    new_position = (position[0] - off[0], position[1] - off[1])
                    img = self.pygame_font.render(
                        text=self.engine.object_to_agent(v.name),
                        antialias=True,
                        color=(0, 0, 0, 128),
                    )
                    # img.set_alpha(None)
                    self.canvas.blit(
                        source=img,
                        dest=(new_position[0] - img.get_width() / 2, new_position[1] - img.get_height() / 2),
                        # special_flags=pygame.BLEND_RGBA_MULT
                    )

    def _handle_event(self) -> None:
        """
        Handle pygame events for moving and zooming in the displayed area.
        """
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    import sys
                    sys.exit()

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()



import logging
import numpy as np
from typing import List, Tuple, Union
from panda3d.core import NodePath
from metadrive.component.lane.abs_lane import AbstractLane

logger = logging.getLogger(__name__)

LaneIndex = Union[str, Tuple[str, str, int]]


class BaseRoadNetwork:
    def __init__(self, debug=False):
        self.graph = None
        self.bounding_box = None
        self._lines_np = []

    def clear(self):
        self.graph.clear()

    def get_bounding_box(self):
        if self.bounding_box is None:
            self.bounding_box = self._get_bounding_box()
        return self.bounding_box

    def _get_bounding_box(self):
        raise NotImplementedError

    def add_lane(self, *args, **kwargs) -> None:
        """
        Add one lane to the roadnetwork for querying
        """
        raise NotImplementedError

    def get_lane(self, index: LaneIndex) -> AbstractLane:
        """
        Get the lane corresponding to a given index in the road network.
        """
        raise NotImplementedError

    def get_closest_lane_index(self, position, return_all=False):
        raise NotImplementedError

    def shortest_path(self, start: str, goal: str) -> List[str]:
        """
        Breadth-first search of shortest checkpoints from start to goal.

        :param start: starting node
        :param goal: goal node
        :return: shortest checkpoints from start to goal.
        """
        raise NotImplementedError

    def __isub__(self, other):
        raise NotImplementedError

    def add(self, other, no_intersect=True):
        """
        Add another network to this one, no intersect means that the same lane should noly exist in self or other
        return: self
        """
        raise NotImplementedError

    def __sub__(self, other):
        ret = self.__class__()
        ret.graph = self.graph
        ret -= other
        return ret

    def show_bounding_box(self, engine, color=None):
        if color is None:
            color = list(np.random.choice(range(256), size=3) / 256) + [1]
        bound_box = self.get_bounding_box()
        points = [(x, y) for x in bound_box[:2] for y in bound_box[2:]]
        for k, p in enumerate(points[:-1]):
            for p_ in points[k + 1:]:
                line = engine.draw_line_3d((*p, 2), (*p_, 2), color, 2)
                line.reparentTo(engine.render)
                self._lines_np.append(line)

    def remove_bounding_box(self):
        if len(self._lines_np) == 0:
            return
        for np in self._lines_np:
            np.detachNode()
            np.removeNode()

    def destroy(self):
        self.bounding_box = None

    def has_connection(self, lane_index_1, lane_index_2):
        """
        Return True if lane 1 is the previous lane of lane 2
        """
        return True if lane_index_2[1] in self.graph[lane_index_1[1]] else False

    def get_map_features(self, interval=2):
        raise NotImplementedError

    def get_all_lanes(self):
        """
        This function will return all lanes in the road network
        :return: list of lanes
        """
        raise NotImplementedError



import logging
from typing import List

from metadrive.utils.math import norm

logger = logging.getLogger(__name__)


class SceneCull:
    """
    Used to cull distant rendering object in MetaDrive to improve rendering efficiency
    """

    # Visualization cull
    LOD_MAP_VIS_DIST = 300  # highly related to the render efficiency !
    LOD_VEHICLE_VIS_DIST = 500
    LOD_OBJECT_VIS_DIST = 500

    # Physics world cull, which can save the time used to do collision detection
    LOD_MAP_PHYSICS_DIST = 50
    LOD_VEHICLE_PHYSICS_DIST = 100000
    LOD_OBJECT_PHYSICS_DIST = 100000

    @classmethod
    def cull_distant_blocks(cls, engine, blocks: list, poses: List[tuple], max_distance=None):
        # A distance based LOD rendering
        for block in blocks:
            if not cls.all_out_of_bounding_box(block.bounding_box, poses, cls.LOD_MAP_VIS_DIST):
                if not block.origin.hasParent():
                    block.origin.reparentTo(engine.worldNP)
            else:
                if block.origin.hasParent():
                    block.origin.detachNode()
            if not cls.all_out_of_bounding_box(block.bounding_box, poses, max_distance or cls.LOD_MAP_PHYSICS_DIST):
                block.dynamic_nodes.attach_to_physics_world(engine.physics_world.dynamic_world)
            else:
                block.dynamic_nodes.detach_from_physics_world(engine.physics_world.dynamic_world)

    @classmethod
    def cull_distant_traffic_vehicles(cls, engine, vehicles: list, poses: List[tuple], max_distance=None):
        cls._cull_elements(
            engine, vehicles, poses, cls.LOD_VEHICLE_VIS_DIST, max_distance or cls.LOD_VEHICLE_PHYSICS_DIST
        )

    @classmethod
    def cull_distant_objects(cls, engine, objects: list, poses: List[tuple], max_distance=None):
        cls._cull_elements(engine, objects, poses, cls.LOD_OBJECT_VIS_DIST, max_distance or cls.LOD_OBJECT_PHYSICS_DIST)

    @classmethod
    def _cull_elements(cls, engine, elements: list, poses: List[tuple], vis_distance: float, physics_distance: float):
        for obj in elements:
            v_p = obj.position
            if not cls.all_distance_greater_than(vis_distance, poses, v_p):
                if not obj.origin.hasParent():
                    obj.origin.reparentTo(engine.pbr_worldNP)
            else:
                if obj.origin.hasParent():
                    obj.origin.detachNode()

            if not cls.all_distance_greater_than(physics_distance, poses, v_p):
                obj.dynamic_nodes.attach_to_physics_world(engine.physics_world.dynamic_world)
            else:
                obj.dynamic_nodes.detach_from_physics_world(engine.physics_world.dynamic_world)

    @staticmethod
    def all_distance_greater_than(distance, poses, target_pos):
        v_p = target_pos
        for pos in poses:
            if norm(v_p[0] - pos[0], v_p[1] - v_p[1]) < distance:
                return False
        return True

    @staticmethod
    def all_out_of_bounding_box(bounding_box, poses, margin_distance):
        for pos in poses:
            if bounding_box[0] - margin_distance < pos[0] < bounding_box[1] + margin_distance and \
                    bounding_box[2] - margin_distance < pos[1] < bounding_box[3] + margin_distance:
                return False
        return True

    @staticmethod
    def out_of_bounding_box(bounding_box, pose, margin_distance):
        bb2 = [
            pose[0] - margin_distance, pose[0] + margin_distance, pose[1] - margin_distance, pose[1] + margin_distance
        ]
        if bounding_box[0] > bb2[1] or \
                bounding_box[1] < bb2[0] or \
                bounding_box[2] > bb2[3] or \
                bounding_box[3] < bb2[2]:
            return True
        else:
            return False

    @staticmethod
    def get_iou(bb1, bb2):
        raise ValueError("not checked")
        """
        Calculate the Intersection over Union (IoU) of two bounding boxes.

        Parameters
        ----------
        bb1 : dict
            Keys: {'x1', 'x2', 'y1', 'y2'}
            The (x1, y1) position is at the top left corner,
            the (x2, y2) position is at the bottom right corner
        bb2 : dict
            Keys: {'x1', 'x2', 'y1', 'y2'}
            The (x, y) position is at the top left corner,
            the (x2, y2) position is at the bottom right corner

        Returns
        -------
        float
            in [0, 1]
        """
        assert bb1['x1'] < bb1['x2']
        assert bb1['y1'] < bb1['y2']
        assert bb2['x1'] < bb2['x2']
        assert bb2['y1'] < bb2['y2']

        # determine the coordinates of the intersection rectangle
        x_left = max(bb1['x1'], bb2['x1'])
        y_top = max(bb1['y1'], bb2['y1'])
        x_right = min(bb1['x2'], bb2['x2'])
        y_bottom = min(bb1['y2'], bb2['y2'])

        if x_right < x_left or y_bottom < y_top:
            return 0.0

        # The intersection of two axis-aligned bounding boxes is always an
        # axis-aligned bounding box
        intersection_area = (x_right - x_left) * (y_bottom - y_top)

        # compute the area of both AABBs
        bb1_area = (bb1['x2'] - bb1['x1']) * (bb1['y2'] - bb1['y1'])
        bb2_area = (bb2['x2'] - bb2['x1']) * (bb2['y2'] - bb2['y1'])

        # compute the intersection over union by taking the intersection
        # area and dividing it by the sum of prediction + ground-truth
        # areas - the interesection area
        iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
        assert iou >= 0.0
        assert iou <= 1.0
        return iou

    @staticmethod
    def no_overlap(bb1, bb2):
        assert bb1['x1'] < bb1['x2']
        assert bb1['y1'] < bb1['y2']
        assert bb2['x1'] < bb2['x2']
        assert bb2['y1'] < bb2['y2']

        # determine the coordinates of the intersection rectangle
        x_left = max(bb1['x1'], bb2['x1'])
        y_top = max(bb1['y1'], bb2['y1'])
        x_right = min(bb1['x2'], bb2['x2'])
        y_bottom = min(bb1['y2'], bb2['y2'])

        if x_right < x_left or y_bottom < y_top:
            return True
        else:
            return False



"""
This file will run environment for one episode and generate two videos:

example_video_RANDOM-STRING/0_bev.mp4
example_video_RANDOM-STRING/0_interface.mp4

This file demonstrates how to use API to generate high-resolution video for demo.
"""
import os
import uuid

import cv2
import matplotlib.pyplot as plt
import pygame

from metadrive.envs import MetaDriveEnv


def plot(d):
    # Can use this function to show images, if you wish to set breakpoints in this file.
    plt.imshow(d)
    plt.show()


class VideoRecorder:
    def __init__(self, video_name, code, height, width, fps=40):
        """
        code=mp4v, avc1, x264, h264 etc.
        """
        assert video_name.endswith(".mp4")
        video = cv2.VideoWriter(video_name, cv2.VideoWriter_fourcc(*code), fps, (width, height))
        self.video_writer = video

    def write(self, image):
        self.video_writer.write(image[..., ::-1])

    def finish(self):
        self.video_writer.release()
        try:
            cv2.destroyAllWindows()
        except Exception as e:
            print("Error happen: ", e)


if __name__ == '__main__':
    num_scenarios = 1  # Will run each seed for 1 episode.
    start_seed = 100
    generate_video = True

    if generate_video:
        folder_name = "example_video_{}".format(str(uuid.uuid4())[:6])

    env = MetaDriveEnv(
        dict(
            num_scenarios=num_scenarios,
            start_seed=start_seed,
            random_traffic=False,
            use_render=True,
            window_size=(1600, 1200),
            crash_vehicle_done=False,
            manual_control=True,  # For using expert policy. You don't need to control it.
        )
    )

    ep_count = 0
    step_count = 0
    frame_count = 0

    o, _ = env.reset(seed=start_seed)
    env.vehicle.expert_takeover = True
    env.engine.force_fps.disable()

    if generate_video:
        os.makedirs(folder_name, exist_ok=True)

        video_base_name = "{}/{}".format(folder_name, ep_count)
        video_name_bev = video_base_name + "_bev.mp4"
        video_bev = VideoRecorder(video_name=video_name_bev, height=3000, width=3000, code="avc1")
        # video_list_bev = []
        video_name_interface = video_base_name + "_interface.mp4"
        # video_list_interface = []
        video_interface = VideoRecorder(video_name=video_name_interface, height=1200, width=1600, code="avc1")

    while True:
        o, r, tm, tc, info = env.step(env.action_space.sample())

        img_interface = env.render("rgb_array")
        img_bev = env.render(
            "topdown",
            target_vehicle_heading_up=False,
            draw_target_vehicle_trajectory=True,
            film_size=(3000, 3000),
            screen_size=(3000, 3000),
            crash_vehicle_done=False,
        )

        if generate_video:
            # Option 1: Directly feed image to video stream. Can save IO time and memory.
            # Has the risk to generate broken video.
            img_bev = pygame.surfarray.array3d(img_bev)
            img_bev = img_bev.swapaxes(0, 1)
            video_bev.write(img_bev)
            video_interface.write(img_interface)

            # Option 2: Save image to local temp folder and generate video later. Waste IO time but save memory.
            # The most safe method.
            # pygame.image.save(img_bev, "{}/{}.png".format(tmp_folder_bev, frame_count))
            # env.capture(file_name="{}/{}.png".format(tmp_folder_interface, frame_count))

            # Note that the interface image can also be used like this:
            # pygame.image.save(img_interface, "{}/{}.png".format(tmp_folder_interface, frame_count))

        frame_count += 1
        step_count += 1

        if tm or tc or step_count > 1000:
            ep_count += 1
            step_count = 0

            env.engine.force_fps.disable()
            if generate_video:
                video_bev.finish()
                video_interface.finish()
                # image_list_to_video(video_name_bev, video_list_bev, code="avc1")
                print("BEV Video is saved at: ", video_name_bev)
                # image_list_to_video(video_name_interface, video_list_interface, code="avc1")
                print("Interface Video is saved at: ", video_name_interface)

                video_base_name = "{}/{}".format(folder_name, ep_count)
                video_name_bev = video_base_name + "_bev.mp4"
                # video_list_bev = []
                video_name_interface = video_base_name + "_interface.mp4"
                # video_list_interface = []

            if ep_count >= num_scenarios:
                break

            if generate_video:
                video_bev = VideoRecorder(video_name=video_name_bev, height=3000, width=3000, code="avc1")
                video_interface = VideoRecorder(video_name=video_name_interface, height=1200, width=1600, code="avc1")

            o, _ = env.reset(seed=ep_count + start_seed)
            env.vehicle.expert_takeover = True
            env.engine.force_fps.disable()

    # if generate_video:
    # from metadrive.utils.image_to_video import image_list_to_video
    # import shutil
    # For Option 2: You need to send the temp folder to video recorder.
    # It will read all image files and generate video.
    # Then delete the temp folder.

    # image_list_to_video(video_name_interface, tmp_folder_interface, code="avc1")
    # print("Interface Video is saved at: ", video_name_interface)
    # shutil.rmtree(tmp_folder_interface)
    #
    # image_list_to_video(video_name_bev, tmp_folder_bev, code="avc1")
    # print("BEV Video is saved at: ", video_name_bev)
    # shutil.rmtree(tmp_folder_bev)



from metadrive import MetaDriveEnv
from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.utils.draw_top_down_map import draw_top_down_map

if __name__ == '__main__':
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            map_config={
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
                BaseMap.GENERATE_CONFIG: "OCrRCTXRCCCCrOr",
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            }
        )
    )
    for i in range(100):
        env.reset()
        map = draw_top_down_map(env.current_map)
        # print("Finish {} maps!".format(i + 1))



import pickle

import tqdm

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import recursive_equal, setup_logger


def test_gen_map_read():
    env_num = 3
    generate_config = {"num_scenarios": env_num, "start_seed": 0}
    restore_config = {"num_scenarios": env_num, "start_seed": 0}

    setup_logger(debug=True)
    try:
        env = MetaDriveEnv(generate_config)
        env.reset()
        data = env.engine.map_manager.dump_all_maps(file_name="test_10maps.pickle")
        env.close()

        # Check load
        with open("test_10maps.pickle", "rb") as f:
            restored_data = pickle.load(f)

        env = MetaDriveEnv(restore_config)
        env.reset()
        env.engine.map_manager.load_all_maps("test_10maps.pickle")

        for i in range(env_num):
            m = env.maps[i + restore_config["start_seed"]].get_meta_data()
            origin = restored_data[i]
            m["map_config"].pop("config")
            m["map_config"].pop("type")

            origin["map_config"].pop("config")
            origin["map_config"].pop("type")

            recursive_equal(m, origin, need_assert=True)
        for seed in tqdm.tqdm(range(env_num), desc="Test Scenario"):
            env.reset(seed=seed)
            for i in range(10):
                env.step(env.action_space.sample())
        # print("Finish!")
    finally:
        env.close()


if __name__ == "__main__":
    test_gen_map_read()



import time

import cupy as cp
import cv2
import numpy as np
import torch
from OpenGL.GL import *  # noqa F403
from cuda import cudart
from cuda.cudart import cudaGraphicsRegisterFlags
from panda3d.core import GraphicsOutput, Texture, GraphicsStateGuardianBase, DisplayRegionDrawCallbackData
from panda3d.core import loadPrcFileData, NodePath
from torch.utils.dlpack import from_dlpack

from metadrive.engine.asset_loader import AssetLoader
from metadrive.tests.vis_block.vis_block_base import TestBlock, BKG_COLOR

# require:
# 1. pip install cupy-cuda12x
# 2. CUDA-Python
# 3. PyOpenGL
# 4. pyrr
# 5. glfw
#


def format_cudart_err(err):
    return (
        f"{cudart.cudaGetErrorName(err)[1].decode('utf-8')}({int(err)}): "
        f"{cudart.cudaGetErrorString(err)[1].decode('utf-8')}"
    )


def check_cudart_err(args):
    if isinstance(args, tuple):
        assert len(args) >= 1
        err = args[0]
        if len(args) == 1:
            ret = None
        elif len(args) == 2:
            ret = args[1]
        else:
            ret = args[1:]
    else:
        err = args
        ret = None

    assert isinstance(err, cudart.cudaError_t), type(err)
    if err != cudart.cudaError_t.cudaSuccess:
        raise RuntimeError(format_cudart_err(err))

    return ret


class CUDATest:
    def __init__(self, window_type="onscreen", shape=None, test_ram_image=False):
        assert shape is not None
        self.engine = engine = TestBlock(window_type=window_type)

        model = self.engine.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
        model.setColor(0.3, 0.5, 0.8)
        model.reparentTo(engine.worldNP)

        self.buffer = self.engine.win.makeTextureBuffer("camera", shape[0], shape[1])

        # this takes care of setting up their camera properly
        self.cam = self.engine.makeCamera(self.buffer)
        self.cam.reparentTo(self.engine.render)

        self.cam.setPos(3, 5, 10)
        self.cam.lookAt(0, 0, 0)

        # buffer property
        self._dtype = np.uint8
        self._shape = shape
        self._strides = None
        self._order = "C"

        self._gl_buffer = None

        self._graphics_resource = None
        self._cuda_buffer = None

        # # make buffer
        # self.texture = self.engine.loader.loadTexture("/home/shady/Desktop/test.jpg")
        self.texture = Texture()
        # self.texture.setMinfilter(Texture.FTLinear)
        # self.texture.setFormat(Texture.FRgba32)
        mode = GraphicsOutput.RTMCopyRam if test_ram_image else GraphicsOutput.RTMBindOrCopy
        self.buffer.addRenderTexture(self.texture, mode)

        def _callback_func(cbdata: DisplayRegionDrawCallbackData):
            cbdata.upcall()
            if not self.registered and self.texture_context_future.done():
                self.register()
            with self as array:
                self.current_data = array

        self.engine.graphicsEngine.renderFrame()
        self.engine.graphicsEngine.renderFrame()
        self.engine.graphicsEngine.renderFrame()
        self.cam.node().getDisplayRegion(0).setDrawCallback(_callback_func)

        self.texture_identifier = None
        self.gsg = GraphicsStateGuardianBase.getDefaultGsg()
        self.texture_context_future = self.texture.prepare(self.gsg.prepared_objects)
        self.new_cuda_mem_ptr = None

        self.current_data = None

    @property
    def cuda_array(self):
        assert self.mapped
        return cp.ndarray(
            shape=(self._shape[1], self._shape[0], self._shape[-1]),
            dtype=self._dtype,
            strides=self._strides,
            order=self._order,
            memptr=self._cuda_buffer
        )

    @property
    def gl_buffer(self):
        return self._gl_buffer

    @property
    def cuda_buffer(self):
        assert self.mapped
        return self._cuda_buffer

    @property
    def graphics_resource(self):
        assert self.registered
        return self._graphics_resource

    @property
    def registered(self):
        return self._graphics_resource is not None

    @property
    def mapped(self):
        return self._cuda_buffer is not None

    def __enter__(self):
        return self.map()

    def __exit__(self, exc_type, exc_value, trace):
        self.unmap()
        return False

    def __del__(self):
        self.unregister()

    def register(self):
        self.texture_identifier = self.texture_context_future.result().getNativeId()
        assert self.texture_identifier is not None
        if self.registered:
            return self._graphics_resource
        self._graphics_resource = check_cudart_err(
            cudart.cudaGraphicsGLRegisterImage(
                self.texture_identifier, GL_TEXTURE_2D, cudaGraphicsRegisterFlags.cudaGraphicsRegisterFlagsReadOnly
            )
        )
        return self._graphics_resource

    def unregister(self):
        if self.registered:
            self.unmap()
            self._graphics_resource = check_cudart_err(cudart.cudaGraphicsUnregisterResource(self._graphics_resource))

    def map(self, stream=0):
        if not self.registered:
            raise RuntimeError("Cannot map an unregistered buffer.")
        if self.mapped:
            return self._cuda_buffer
        # self.engine.graphicsEngine.renderFrame()
        check_cudart_err(cudart.cudaGraphicsMapResources(1, self._graphics_resource, stream))
        array = check_cudart_err(cudart.cudaGraphicsSubResourceGetMappedArray(self.graphics_resource, 0, 0))
        channelformat, cudaextent, flag = check_cudart_err(cudart.cudaArrayGetInfo(array))

        depth = 1
        byte = 4  # four channel
        if self.new_cuda_mem_ptr is None:
            success, self.new_cuda_mem_ptr = cudart.cudaMalloc(cudaextent.height * cudaextent.width * byte * depth)
        check_cudart_err(
            cudart.cudaMemcpy2DFromArray(
                self.new_cuda_mem_ptr, cudaextent.width * byte * depth, array, 0, 0, cudaextent.width * byte * depth,
                cudaextent.height, cudart.cudaMemcpyKind.cudaMemcpyDeviceToDevice
            )
        )
        if self._cuda_buffer is None:
            self._cuda_buffer = cp.cuda.MemoryPointer(
                cp.cuda.UnownedMemory(self.new_cuda_mem_ptr, cudaextent.width * depth * byte * cudaextent.height, self),
                0
            )
        return self.cuda_array

    def unmap(self, stream=None):
        if not self.registered:
            raise RuntimeError("Cannot unmap an unregistered buffer.")
        if not self.mapped:
            return self

        self._cuda_buffer = check_cudart_err(cudart.cudaGraphicsUnmapResources(1, self._graphics_resource, stream))

        return self

    def step(self):
        self.engine.taskMgr.step()


if __name__ == "__main__":
    win_size = (1200, 800)
    loadPrcFileData("", "threading-model Cull/Draw")
    loadPrcFileData("", "textures-power-2 none")
    loadPrcFileData("", "win-size {} {}".format(*win_size))
    test_ram_image = False
    render = True
    env = CUDATest(window_type="offscreen", shape=(*win_size, 4), test_ram_image=test_ram_image)

    # important
    env.engine.graphicsEngine.renderFrame()
    env.engine.graphicsEngine.renderFrame()

    start = time.time()
    for s in range(10000000):
        env.step()
        if test_ram_image:
            origin_img = env.texture
            img = np.frombuffer(origin_img.getRamImage().getData(), dtype=np.uint8)
            img = img.reshape((origin_img.getYSize(), origin_img.getXSize(), 4))
            img = img
            torch_img = torch.from_numpy(img)
            if render:
                cv2.imshow("win", img)
                cv2.waitKey(1)
        else:
            ret = from_dlpack(env.current_data.toDlpack())
            if render:
                np_array = cp.asnumpy(ret)[::-1]
                cv2.imshow("win", np_array)
                cv2.waitKey(1)
        if s % 10000 == 0 and s != 0:
            print("FPS: {}".format(s / (time.time() - start)))



import numpy as np

from metadrive.component.vehicle_model.bicycle_model import BicycleModel
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger
from metadrive.utils.math import norm


def predict(current_state, actions, model):
    model.reset(*current_state)
    for action in actions:
        model.predict(0.1, action)
    return model.state


def _test_bicycle_model():
    horizon = 10
    setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": .0,
            "use_render": True,
            # "manual_control": True,
            "map": "CCCC",
            "vehicle_config": {
                "enable_reverse": False,
            }
        }
    )
    bicycle_model = BicycleModel()
    o, _ = env.reset()
    vehicle = env.current_track_vehicle
    v_dir = vehicle.velocity_direction
    bicycle_model.reset(*vehicle.position, vehicle.speed, vehicle.heading_theta, np.arctan2(v_dir[1], v_dir[0]))
    actions = []
    for steering in [1.0, 0.8, 0.6, 0.4, 0.2, 0]:
        for dir in [-1, 1]:
            s = dir * steering
            for throttle in [1.0, 0.8, 0.6, 0.4, 0.2, 0, -0.5]:
                actions += [[s, throttle]] * 20
    predict_states = []
    for s in range(len(actions)):
        vehicle = env.current_track_vehicle
        v_dir = vehicle.velocity_direction
        predict_states.append(
            predict(
                current_state=(
                    *env.current_track_vehicle.position, env.current_track_vehicle.speed,
                    env.current_track_vehicle.heading_theta, np.arctan2(v_dir[1], v_dir[0])
                ),
                actions=[actions[i] for i in range(s, s + horizon)],
                model=bicycle_model
            )
        )
        o, r, tm, tc, info = env.step(actions[s])
        index = s - horizon
        if index >= 0:
            state = predict_states[index]
            # print(norm(state["x"] - vehicle.position[0], state["y"] - vehicle.position[1]))


if __name__ == "__main__":
    _test_bicycle_model()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import os
import atexit

from panda3d.core import Filename, VirtualFileSystem, get_model_path
from panda3d.core import VirtualFileMountRamdisk
from direct.stdpy.file import join, isdir, isfile

from metadrive.render_pipeline.rpcore.rpobject import RPObject


class MountManager(RPObject):
    """ This classes mounts the required directories for the pipeline to run.
    This is important if the pipeline is in a subdirectory for example. The
    mount manager also handles the lock, storing the current PID into a file
    named instance.pid and ensuring that there is only 1 instance of the
    pipeline running at one time. """
    def __init__(self, pipeline):
        """ Creates a new mount manager """
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._base_path = self._find_basepath()
        self._lock_file = "instance.pid"
        self._model_paths = []
        self._write_path = None
        self._mounted = False
        self._do_cleanup = True
        self._config_dir = None

        self.debug("Auto-Detected base path to", self._base_path)
        atexit.register(self._on_exit_cleanup)

    @property
    def write_path(self):
        """ Returns the write path previously set with set_write_path, or None
        if no write path has been set yet. """
        return self._write_path

    @write_path.setter
    def write_path(self, pth):
        """ Set a writable directory for generated files. This can be a string
        path name or a multifile with openReadWrite(). If no pathname is set
        then the root directory is used.

        This feature is usually only used for debugging, the pipeline will dump
        all generated shaders and other temporary files to that directory.
        If you don't need this, you can use set_virtual_write_path(), which
        will create the temporary path in the VirtualFileSystem, thus not
        writing any files to disk. """
        if pth is None:
            self._write_path = None
            self._lock_file = "instance.pid"
        else:
            self._write_path = Filename.from_os_specific(pth).get_fullpath()
            self._lock_file = join(self._write_path, "instance.pid")

    @property
    def base_path(self):
        """ Returns the base path of the pipeline. This returns the path previously
        set with set_base_path, or the auto detected base path if no path was
        set yet """
        return self._base_path

    @base_path.setter
    def base_path(self, pth):
        """ Sets the path where the base shaders and models on are contained. This
        is usually the root of the rendering pipeline folder """
        self.debug("Set base path to '" + pth + "'")
        self._base_path = Filename.from_os_specific(pth).get_fullpath()

    @property
    def config_dir(self):
        """ Returns the config directory previously set with set_config_dir, or
        None if no directory was set yet """

    @config_dir.setter
    def config_dir(self, pth):
        """ Sets the path to the config directory. Usually this is the config/
        directory located in the pipeline root directory. However, if you want
        to load your own configuration files, you can specify a custom config
        directory here. Your configuration directory should contain the
        pipeline.yaml, plugins.yaml, daytime.yaml and configuration.prc.

        It is highly recommended you use the pipeline provided config files, modify
        them to your needs, and as soon as you think they are in a final version,
        copy them over. Please also notice that you should keep your config files
        up-to-date, e.g. when new configuration variables are added.

        Also, specifying a custom configuration_dir disables the functionality
        of the PluginConfigurator and DayTime editor, since they operate on the
        pipelines default config files.

        Set the directory to None to use the default directory. """
        self._config_dir = Filename.from_os_specific(pth).get_fullpath()

    @property
    def do_cleanup(self):
        """ Returns whether the mount manager will attempt to cleanup the
        generated files after the application stopped running """
        return self._do_cleanup

    @do_cleanup.setter
    def do_cleanup(self, cleanup):
        """ Sets whether to cleanup the tempfolder after the application stopped.
        This is mostly useful for debugging, to analyze the generated tempfiles
        even after the pipeline stopped running """
        self._do_cleanup = cleanup

    def get_lock(self):
        """ Checks if we are the only instance running. If there is no instance
        running, write the current PID to the instance.pid file. If the
        instance file exists, checks if the specified process still runs. This
        way only 1 instance of the pipeline can be run at one time. """

        # Check if there is a lockfile at all
        if isfile(self._lock_file):
            # Read process id from lockfile
            try:
                with open(self._lock_file, "r") as handle:
                    pid = int(handle.readline())
            except IOError as msg:
                self.error("Failed to read lockfile:", msg)
                return False

            # Check if the process is still running
            if self._is_pid_running(pid):
                self.error("Found running instance")
                return False

            # Process is not running anymore, we can write the lockfile
            self._write_lock()
            return True

        else:
            # When there is no lockfile, just create it and continue
            self._write_lock()
            return True

    def _find_basepath(self):
        """ Attempts to find the pipeline base path by looking at the location
        of this file """
        pth = os.path.abspath(join(os.path.dirname(os.path.realpath(__file__)), ".."))
        return Filename.from_os_specific(pth).get_fullpath()

    def _is_pid_running(self, pid):
        """ Checks if a pid is still running """

        # Code snippet from:
        # http://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid

        if os.name == 'posix':
            import errno
            if pid < 0:
                return False
            try:
                os.kill(pid, 0)
            except OSError as err:
                return err.errno == errno.EPERM
            else:
                return True
        else:
            import ctypes
            kernel32 = ctypes.windll.kernel32
            process = kernel32.OpenProcess(0x100000, 0, pid)
            if process != 0:
                kernel32.CloseHandle(process)
                return True
            else:
                return False

    def _write_lock(self):
        """ Internal method to write the current process id to the instance.pid
        lockfile. This is used to ensure no second instance of the pipeline is
        running. """

        with open(self._lock_file, "w") as handle:
            handle.write(str(os.getpid()))

    def _try_remove(self, fname):
        """ Tries to remove the specified filename, returns either True or False
        depending if we had success or not """
        try:
            os.remove(fname)
            return True
        except (IOError, OSError):
            pass
        return False

    def _on_exit_cleanup(self):
        """ Gets called when the application exists """

        if self._do_cleanup:
            self.debug("Cleaning up ..")

            if self._write_path is not None:

                # Try removing the lockfile
                self._try_remove(self._lock_file)

                # Check for further tempfiles in the write path
                # We explicitely use os.listdir here instead of pandas listdir,
                # to work with actual paths
                for fname in os.listdir(self._write_path):
                    pth = join(self._write_path, fname)

                    # Tempfiles from the pipeline start with "$$" to distinguish
                    # them from user created files
                    if isfile(pth) and fname.startswith("$$"):
                        self._try_remove(pth)

                # Delete the write path if no files are left
                if len(os.listdir(self._write_path)) < 1:
                    try:
                        os.removedirs(self._write_path)
                    except IOError:
                        pass

    @property
    def is_mounted(self):
        """ Returns whether the MountManager was already mounted by calling
        mount() """
        return self._mounted

    def mount(self):
        """ Inits the VFS Mounts. This creates the following virtual directory
        structure, from which all files can be located:

        /$$rp/  (Mounted from the render pipeline base directory)
           + rpcore/
           + shader/
           + ...

        /$rpconfig/ (Mounted from config/, may be set by user)
           + pipeline.yaml
           + ...

        /$$rptemp/ (Either ramdisk or user specified)
            + day_time_config
            + shader_auto_config
            + ...

        /$$rpshader/ (Link to /$$rp/rpcore/shader)

         """
        self.debug("Setting up virtual filesystem")
        self._mounted = True

        def convert_path(pth):
            return Filename.from_os_specific(pth).get_fullpath()

        vfs = VirtualFileSystem.get_global_ptr()

        # Mount config dir as $$rpconf
        if self._config_dir is None:
            config_dir = convert_path(join(self._base_path, "config/"))
            self.debug("Mounting auto-detected config dir:", config_dir)
            vfs.mount(config_dir, "/$$rpconfig", 0)
        else:
            self.debug("Mounting custom config dir:", self._config_dir)
            vfs.mount(convert_path(self._config_dir), "/$$rpconfig", 0)

        # Mount directory structure
        vfs.mount(convert_path(self._base_path), "/$$rp", 0)
        vfs.mount(convert_path(join(self._base_path, "rpcore/shader")), "/$$rp/shader", 0)
        vfs.mount(convert_path(join(self._base_path, "effects")), "effects", 0)

        # Mount the pipeline temp path:
        # If no write path is specified, use a virtual ramdisk
        if self._write_path is None:
            self.debug("Mounting ramdisk as /$$rptemp")
            vfs.mount(VirtualFileMountRamdisk(), "/$$rptemp", 0)
        else:
            # In case an actual write path is specified:
            # Ensure the pipeline write path exists, and if not, create it
            if not isdir(self._write_path):
                self.debug("Creating temporary path, since it does not exist yet")
                try:
                    os.makedirs(self._write_path)
                except IOError as msg:
                    self.fatal("Failed to create temporary path:", msg)
            self.debug("Mounting", self._write_path, "as /$$rptemp")
            vfs.mount(convert_path(self._write_path), '/$$rptemp', 0)

        get_model_path().prepend_directory("/$$rp")
        get_model_path().prepend_directory("/$$rp/shader")
        get_model_path().prepend_directory("/$$rptemp")

    def unmount(self):
        """ Unmounts the VFS """
        raise NotImplementedError("TODO")



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from __future__ import print_function

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error
from panda3d.core import PTAVecBase3f

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.pluginbase.base_plugin import BasePlugin
from metadrive.render_pipeline.rpcore.native import PSSMCameraRig, NATIVE_CXX_LOADED

from .pssm_stage import PSSMStage
from .pssm_shadow_stage import PSSMShadowStage
from .pssm_scene_shadow_stage import PSSMSceneShadowStage
from .pssm_dist_shadow_stage import PSSMDistShadowStage


class Plugin(BasePlugin):

    name = "Sun Shadows"
    author = "tobspr <tobias.springer1@gmail.com>"
    description = ("This plugin adds support for Parallel Split Shadow Maps " "(PSSM), and also sun lighting.")
    version = "1.2"
    required_plugins = ("scattering", )

    def on_stage_setup(self):

        if not NATIVE_CXX_LOADED:
            self.debug("Setting max splits to 1 since python is used")
            self._pipeline.plugin_mgr.settings["pssm"]["split_count"].set_value(1)

        self.update_enabled = True
        self.pta_sun_vector = PTAVecBase3f.empty_array(1)
        self.last_cache_reset = 0

        self.shadow_stage = self.create_stage(PSSMShadowStage)
        self.pssm_stage = self.create_stage(PSSMStage)

        self.shadow_stage.num_splits = self.get_setting("split_count")
        self.shadow_stage.split_resolution = self.get_setting("resolution")

        self.scene_shadow_stage = self.create_stage(PSSMSceneShadowStage)
        self.scene_shadow_stage.resolution = self.get_setting("scene_shadow_resolution")
        self.scene_shadow_stage.sun_distance = self.get_setting("scene_shadow_sundist")

        # Enable distant shadow map if specified
        if self.get_setting("use_distant_shadows"):
            self.dist_shadow_stage = self.create_stage(PSSMDistShadowStage)
            self.dist_shadow_stage.resolution = self.get_setting("dist_shadow_resolution")
            self.dist_shadow_stage.clip_size = self.get_setting("dist_shadow_clipsize")
            self.dist_shadow_stage.sun_distance = self.get_setting("dist_shadow_sundist")

            self.pssm_stage.required_pipes.append("PSSMDistSunShadowMap")
            self.pssm_stage.required_inputs.append("PSSMDistSunShadowMapMVP")

    def on_pipeline_created(self):
        self.debug("Initializing pssm ..")
        # Construct a dummy node to parent the rig to
        self.node = Globals.base.render.attach_new_node("PSSMCameraRig")
        self.node.hide()

        # Construct the actual PSSM rig
        self.camera_rig = PSSMCameraRig(self.get_setting("split_count"))  # noqa # pylint: disable=undefined-variable
        self.camera_rig.set_sun_distance(self.get_setting("sun_distance"))
        self.camera_rig.set_pssm_distance(self.get_setting("max_distance"))
        self.camera_rig.set_logarithmic_factor(self.get_setting("logarithmic_factor"))
        self.camera_rig.set_border_bias(self.get_setting("border_bias"))
        self.camera_rig.set_use_stable_csm(True)
        self.camera_rig.set_use_fixed_film_size(True)
        self.camera_rig.set_resolution(self.get_setting("resolution"))
        self.camera_rig.reparent_to(self.node)

        # Attach the cameras to the shadow stage
        for i in range(self.get_setting("split_count")):
            camera_np = self.camera_rig.get_camera(i)
            camera_np.node().set_scene(Globals.base.render)
            region = self.shadow_stage.split_regions[i]
            region.set_camera(camera_np)
            # camera_np.node().show_frustum()

            # Make sure the pipeline knows about our camera, so it can apply
            # the correct bitmasks
            self._pipeline.tag_mgr.register_camera("shadow", camera_np.node())

        # Accept a shortcut to enable / disable the update of PSSM
        Globals.base.accept("u", self.toggle_update_enabled)

        # Set inputs
        self.pssm_stage.set_shader_inputs(
            pssm_mvps=self.camera_rig.get_mvp_array(), pssm_nearfar=self.camera_rig.get_nearfar_array()
        )

        if self.is_plugin_enabled("volumetrics"):
            handle = self.get_plugin_instance("volumetrics")
            handle.stage.set_shader_inputs(
                pssm_mvps=self.camera_rig.get_mvp_array(), pssm_nearfar=self.camera_rig.get_nearfar_array()
            )

    def on_pre_render_update(self):
        sun_vector = self.get_plugin_instance("scattering").sun_vector

        if sun_vector.z < 0.0:
            self.shadow_stage.active = False
            self.scene_shadow_stage.active = False
            self.pssm_stage.set_render_shadows(False)

            if self.get_setting("use_distant_shadows"):
                self.dist_shadow_stage.active = False

            # Return, no need to update the pssm splits
            return
        else:
            self.shadow_stage.active = True
            self.scene_shadow_stage.active = True
            self.pssm_stage.set_render_shadows(True)

            if self.get_setting("use_distant_shadows"):
                self.dist_shadow_stage.active = True

        if self.update_enabled:
            self.camera_rig.update(Globals.base.camera, sun_vector)

            # Eventually reset cache
            cache_diff = Globals.clock.get_frame_time() - self.last_cache_reset
            if cache_diff > 5.0:
                self.last_cache_reset = Globals.clock.get_frame_time()
                self.camera_rig.reset_film_size_cache()

            self.scene_shadow_stage.sun_vector = sun_vector

            if self.get_setting("use_distant_shadows"):
                self.dist_shadow_stage.sun_vector = sun_vector

    def update_max_distance(self):
        self.camera_rig.set_pssm_distance(self.get_setting("max_distance"))

    def update_logarithmic_factor(self):
        self.camera_rig.set_logarithmic_factor(self.get_setting("logarithmic_factor"))

    def update_border_bias(self):
        self.camera_rig.set_border_bias(self.get_setting("border_bias"))

    def update_sun_distance(self):
        self.camera_rig.set_sun_distance(self.get_setting("sun_distance"))

    def update_dist_shadow_clipsize(self):
        self.dist_shadow_stage.clip_size = self.get_setting("dist_shadow_clipsize")

    def update_dist_shadow_sundist(self):
        self.dist_shadow_stage.sun_distance = self.get_setting("dist_shadow_sundist")

    def update_scene_shadow_sundist(self):
        self.scene_shadow_stage.sun_distance = self.get_setting("scene_shadow_sundist")

    def toggle_update_enabled(self):
        self.update_enabled = not self.update_enabled
        self.debug("Update enabled:", self.update_enabled)



try:
    from OpenGL.GL import *  # noqa F403
    import cupy as cp
    from cuda import cudart
except ImportError:
    pass


def format_cudart_err(err):
    return (
        f"{cudart.cudaGetErrorName(err)[1].decode('utf-8')}({int(err)}): "
        f"{cudart.cudaGetErrorString(err)[1].decode('utf-8')}"
    )


def check_cudart_err(args):
    if isinstance(args, tuple):
        assert len(args) >= 1
        err = args[0]
        if len(args) == 1:
            ret = None
        elif len(args) == 2:
            ret = args[1]
        else:
            ret = args[1:]
    else:
        err = args
        ret = None

    assert isinstance(err, cudart.cudaError_t), type(err)
    if err != cudart.cudaError_t.cudaSuccess:
        raise RuntimeError(format_cudart_err(err))

    return ret


def restore_channel_and_shape(cuda_image):
    raise DeprecationWarning()
    return cuda_image[..., ::-1][::-1]



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv
from metadrive.manager.traffic_manager import TrafficMode
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import setup_logger
from metadrive.utils.math import norm


def assert_equal_pos(dict_1, dict_2):
    for name, pos in dict_1.items():
        assert name in dict_2, "No such vehicle in dict 2: {}".format(name)
        pos_2 = dict_2[name]
        difference = pos - pos_2
        diff = norm(difference[0], difference[1])
        assert diff < 1, "pos mismatch for vehicle: {}, distance: {}".format(name, diff)


def test_save_recreate_scenario(vis=False):
    setup_logger(True)
    cfg = {
        "accident_prob": 0.8,
        "num_scenarios": 1,
        "traffic_density": 0.1,
        "start_seed": 1000,
        "manual_control": False,
        "debug": False,
        "use_render": vis,
        "agent_policy": IDMPolicy,
        "traffic_mode": TrafficMode.Trigger,
        "record_episode": True,
        "map_config": {
            BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
            BaseMap.GENERATE_CONFIG: "CrCSX",
            BaseMap.LANE_WIDTH: 3.5,
            BaseMap.LANE_NUM: 3,
        }
    }
    env = SafeMetaDriveEnv(cfg)
    try:
        positions_1 = []
        o, _ = env.reset()
        epi_info = env.engine.record_manager.get_episode_metadata()
        for i in range(1, 100000 if vis else 2000):
            o, r, tm, tc, info = env.step([0, 1])
            positions_1.append({v.name: v.position for v in env.engine.traffic_manager.spawned_objects.values()})
            if tm or tc:
                break
        env.close()
        env = SafeMetaDriveEnv(cfg)
        env.config["replay_episode"] = epi_info
        env.config["record_episode"] = False
        env.config["only_reset_when_replay"] = True
        o, _ = env.reset()
        positions_1.reverse()
        for i in range(0, 100000 if vis else 2000):
            o, r, tm, tc, info = env.step([0, 1])
            position = positions_1.pop()
            position = {env.engine.replay_manager.record_name_to_current_name[key]: v for key, v in position.items()}
            current_position = {v.name: v.position for v in env.engine.traffic_manager.spawned_objects.values()}
            assert_equal_pos(position, current_position)
            if tm or tc:
                break
    finally:
        env.close()


if __name__ == "__main__":
    test_save_recreate_scenario(vis=False)



"""
Similar to native MetaDriveEnv, which allows PG map generation to create infinite number of scenarios for
generalization experiments, this file provides a environment where you can further randomize the dynamics of ego
vehicle.

Note that the sampled dynamics parameters will not be changed if you don't change the global seed.
This means that if num_scenarios = 1, then you will deterministically sample the same agent with
the same dynamics. Set num_scenarios > 1 to allow more diverse dynamics.
"""

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.manager.agent_manager import AgentManager

VaryingDynamicsConfig = dict(
    vehicle_config=dict(vehicle_model="varying_dynamics", ),
    random_dynamics=dict(
        # We will sample each parameter from (min_value, max_value)
        # You can set it to None to stop randomizing the parameter.
        max_engine_force=(100, 3000),
        max_brake_force=(20, 600),
        wheel_friction=(0.1, 2.5),
        max_steering=(10, 80),  # The maximum steering angle if action = +-1
        mass=(300, 3000)
    )
)


class VaryingDynamicsAgentManager(AgentManager):
    def reset(self):
        # Randomize ego vehicle's dynamics here
        random_fields = self.engine.global_config["random_dynamics"]
        dynamics = {}
        for parameter, para_range in random_fields.items():
            if para_range is None:
                continue
            elif isinstance(para_range, (tuple, list)):
                assert len(para_range) == 2
                assert para_range[1] >= para_range[0]
                if para_range[1] == para_range[0]:
                    dynamics[parameter] = para_range[0]
                else:
                    dynamics[parameter] = self.np_random.uniform(para_range[0], para_range[1])
            else:
                raise ValueError("Unknown parameter range: {}".format(para_range))

        assert len(self.engine.global_config["target_vehicle_configs"]) == 1, "Only supporting single-agent now!"
        self.engine.global_config["target_vehicle_configs"]["default_agent"].update(dynamics)
        super(VaryingDynamicsAgentManager, self).reset()


class VaryingDynamicsEnv(MetaDriveEnv):
    @classmethod
    def default_config(cls):
        config = super(VaryingDynamicsEnv, cls).default_config()
        config.update(VaryingDynamicsConfig)
        return config

    def _get_agent_manager(self):
        return VaryingDynamicsAgentManager(
            init_observations=self._get_observations(), init_action_space=self._get_action_space()
        )


if __name__ == '__main__':
    # Local test
    env = VaryingDynamicsEnv({
        "num_scenarios": 10,  # Allow 10 random envs.
    })
    for ep in range(3):
        obs, _ = env.reset()
        print("Current Dynamics Parameters:", env.vehicle.get_dynamics_parameters())
        for step in range(1000):
            o, r, tm, tc, i = env.step(env.action_space.sample())
            if tm or tc:
                break



from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv
from metadrive.utils import setup_logger

if __name__ == "__main__":
    setup_logger(True)
    env = SafeMetaDriveEnv(
        {
            "use_render": True,
            "manual_control": True,
            "num_scenarios": 100,
            "show_coordinates": True,
            "map": "SSSS",
            "accident_prob": 1.0,
            "vehicle_config": {
                "show_lidar": True
            }
        }
    )

    o, _ = env.reset()
    # print("vehicle num", len(env.engine.traffic_manager.vehicles))
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        env.render(text={})
    env.close()



"""

Example of using the render service to render a simple sphere.
The render service should be running.

"""

import sys
import pickle
import socket

# Address of the render service
renderer_ip = ("127.0.0.1", 62360)

# Port which the render service will ping when its done.
pingback_port = 62500

# Scene data
payload = pickle.dumps(
    {
        "scene": "resources/preview.bam",
        "dest": "demo-render.png",
        "view_size_x": 512,
        "view_size_y": 512,
        "pingback_port": pingback_port
    },
    protocol=0
)

# Start rendering
print("Sending payload ...")
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
try:
    sock.sendto(payload, renderer_ip)
except Exception as msg:
    sock.close()
    raise msg

# Wait until the renderer finished
print("Waiting for response ...")
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
sock.settimeout(3.0)

try:
    sock.bind(("localhost", pingback_port))
except Exception as msg:
    print("Could not connect to pingback_port - maybe a render result is still waiting?")
    print("Error message:", msg)
    sys.exit(1)

sock.listen(1)

try:
    conn, addr = sock.accept()
except socket.timeout:
    print("No render pipeline response within timeout! Make sure the service is running.")
    sock.close()
    sys.exit(1)

print("Rendering was done!")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main_window.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(973, 632)
        MainWindow.setMinimumSize(QtCore.QSize(973, 632))
        MainWindow.setMaximumSize(QtCore.QSize(973, 632))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(63, 63, 63))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(52, 52, 52))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(28, 28, 28))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ToolTipText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(63, 63, 63))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(52, 52, 52))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(28, 28, 28))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ToolTipText, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(63, 63, 63))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(52, 52, 52))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(28, 28, 28))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(42, 42, 42))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ToolTipText, brush)
        MainWindow.setPalette(palette)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/res/res/icon.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        MainWindow.setAutoFillBackground(False)
        MainWindow.setStyleSheet(
            "QMainWindow { background: #fff;}\n"
            "*, QLabel { font-family: Roboto; font-weight: 300; }\n"
            "\n"
            "QScrollBar {\n"
            "    background: #eee;\n"
            "}\n"
            "\n"
            "\n"
            "QScrollBar:vertical {\n"
            "                        width: 9px;\n"
            "    margin: 0;\n"
            "                      }\n"
            "\n"
            "                      QScrollBar::handle:vertical {\n"
            "                        min-height: 15px;\n"
            "                        background: #aaa;\n"
            "\n"
            "                      }\n"
            "\n"
            "\n"
            "                      QScrollBar::handle:vertical:hover {\n"
            "                        \n"
            "                        background: #999;\n"
            "\n"
            "                      }\n"
            "                      QScrollBar::add-line:vertical {\n"
            "                      }\n"
            "\n"
            "                      QScrollBar::sub-line:vertical {\n"
            "                      }\n"
            "\n"
            "\n"
            "                      QScrollBar::add-page:vertical {\n"
            "                        background: #ddd;\n"
            "\n"
            "                      }\n"
            "                      QScrollBar::sub-page:vertical {\n"
            "                        background: #ddd;\n"
            "                    }\n"
            ""
        )
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.frame_current_setting = QtWidgets.QFrame(self.centralwidget)
        self.frame_current_setting.setGeometry(QtCore.QRect(350, 80, 641, 551))
        self.frame_current_setting.setStyleSheet("QFrame { background: #e5e5e5;}")
        self.frame_current_setting.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_current_setting.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_current_setting.setObjectName("frame_current_setting")
        self.lbl_current_setting = QtWidgets.QLabel(self.frame_current_setting)
        self.lbl_current_setting.setGeometry(QtCore.QRect(20, 16, 221, 31))
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setPointSize(-1)
        font.setBold(False)
        font.setWeight(37)
        self.lbl_current_setting.setFont(font)
        self.lbl_current_setting.setStyleSheet("color: #555; border: 0;  font-size: 15px; text-transform: uppercase;")
        self.lbl_current_setting.setObjectName("lbl_current_setting")
        self.btn_reset = QtWidgets.QPushButton(self.frame_current_setting)
        self.btn_reset.setGeometry(QtCore.QRect(500, 14, 101, 31))
        self.btn_reset.setStyleSheet(
            "QPushButton {\n"
            "color: #eee;\n"
            "background: #666;\n"
            "border: 0; \n"
            "}\n"
            "\n"
            "QPushButton:hover {\n"
            "    background: #555;\n"
            "}\n"
            "\n"
            "QPushButton:pressed {\n"
            "    background: #111;\n"
            "}"
        )
        self.btn_reset.setObjectName("btn_reset")
        self.frame_3 = QtWidgets.QFrame(self.frame_current_setting)
        self.frame_3.setGeometry(QtCore.QRect(20, 94, 581, 441))
        self.frame_3.setStyleSheet("QFrame {\n" "}")
        self.frame_3.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.frame_3)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(0, 0, 571, 441))
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.prefab_edit_widget = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.prefab_edit_widget.setContentsMargins(0, 0, 0, 0)
        self.prefab_edit_widget.setSpacing(0)
        self.prefab_edit_widget.setObjectName("prefab_edit_widget")
        self.lbl_setting_desc = QtWidgets.QLabel(self.frame_current_setting)
        self.lbl_setting_desc.setGeometry(QtCore.QRect(20, 50, 571, 41))
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setBold(False)
        font.setWeight(37)
        self.lbl_setting_desc.setFont(font)
        self.lbl_setting_desc.setStyleSheet("color: #888;")
        self.lbl_setting_desc.setWordWrap(True)
        self.lbl_setting_desc.setObjectName("lbl_setting_desc")
        self.btn_insert_point = QtWidgets.QPushButton(self.frame_current_setting)
        self.btn_insert_point.setGeometry(QtCore.QRect(368, 14, 131, 31))
        self.btn_insert_point.setStyleSheet(
            "QPushButton {\n"
            "color: #eee;\n"
            "background: #666;\n"
            "border: 0; \n"
            "}\n"
            "\n"
            "QPushButton:hover {\n"
            "    background: #555;\n"
            "}\n"
            "\n"
            "QPushButton:pressed {\n"
            "    background: #111;\n"
            "}"
        )
        self.btn_insert_point.setObjectName("btn_insert_point")
        self.frame_2 = QtWidgets.QFrame(self.centralwidget)
        self.frame_2.setGeometry(QtCore.QRect(-1, 80, 351, 551))
        self.frame_2.setStyleSheet("background: #eee;")
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.settings_tree = QtWidgets.QTreeWidget(self.frame_2)
        self.settings_tree.setGeometry(QtCore.QRect(14, 20, 311, 511))
        self.settings_tree.setStyleSheet(
            "QTreeWidget::item {\n"
            "padding: 7px 7px 6px;\n"
            "background: #ddd;\n"
            "outline: 0 !important;\n"
            "margin-bottom: 1px;\n"
            "margin-right: 0px;\n"
            "color: #777;\n"
            "border: 0;\n"
            "border-radius: 0;\n"
            "}\n"
            "\n"
            "QTreeWidget::item:hover {\n"
            "background: #ccc;\n"
            "}\n"
            "\n"
            "QTreeWidget::item:selected {\n"
            "background: #555;\n"
            "color: #eee;\n"
            "}\n"
            "\n"
            "QTreeWidget {\n"
            "padding: 0;\n"
            "color: #eee;\n"
            "background: transparent;\n"
            "\n"
            "}\n"
            "\n"
            "* {\n"
            "outline: 0;\n"
            "}\n"
            "\n"
            ""
        )
        self.settings_tree.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.settings_tree.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.settings_tree.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.settings_tree.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.settings_tree.setProperty("showDropIndicator", False)
        self.settings_tree.setAlternatingRowColors(False)
        self.settings_tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.settings_tree.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.settings_tree.setIndentation(18)
        self.settings_tree.setRootIsDecorated(True)
        self.settings_tree.setUniformRowHeights(False)
        self.settings_tree.setItemsExpandable(True)
        self.settings_tree.setAnimated(True)
        self.settings_tree.setAllColumnsShowFocus(False)
        self.settings_tree.setWordWrap(True)
        self.settings_tree.setHeaderHidden(True)
        self.settings_tree.setObjectName("settings_tree")
        item_0 = QtWidgets.QTreeWidgetItem(self.settings_tree)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_0 = QtWidgets.QTreeWidgetItem(self.settings_tree)
        item_0.setFlags(
            QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsUserCheckable
            | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsTristate
        )
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.settings_tree.header().setVisible(False)
        self.settings_tree.header().setCascadingSectionResizes(False)
        self.settings_tree.header().setDefaultSectionSize(150)
        self.settings_tree.header().setStretchLastSection(False)
        self.frame_4 = QtWidgets.QFrame(self.centralwidget)
        self.frame_4.setGeometry(QtCore.QRect(420, 10, 551, 61))
        self.frame_4.setStyleSheet("QFrame {\n" "    background: #fff;\n" "}\n" "")
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_4.setObjectName("frame_4")
        self.time_slider = QtWidgets.QSlider(self.frame_4)
        self.time_slider.setGeometry(QtCore.QRect(10, 11, 521, 20))
        self.time_slider.setMaximum(5184000)
        self.time_slider.setSingleStep(3600)
        self.time_slider.setPageStep(3600)
        self.time_slider.setProperty("value", 2592000)
        self.time_slider.setTracking(True)
        self.time_slider.setOrientation(QtCore.Qt.Horizontal)
        self.time_slider.setInvertedAppearance(False)
        self.time_slider.setInvertedControls(False)
        self.time_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.time_slider.setTickInterval(432000)
        self.time_slider.setObjectName("time_slider")
        self.label_2 = QtWidgets.QLabel(self.frame_4)
        self.label_2.setGeometry(QtCore.QRect(254, 30, 32, 32))
        self.label_2.setText("")
        self.label_2.setPixmap(QtGui.QPixmap(":/res/res/12_00.png"))
        self.label_2.setObjectName("label_2")
        self.label_3 = QtWidgets.QLabel(self.frame_4)
        self.label_3.setGeometry(QtCore.QRect(341, 30, 32, 32))
        self.label_3.setText("")
        self.label_3.setPixmap(QtGui.QPixmap(":/res/res/16_00.png"))
        self.label_3.setObjectName("label_3")
        self.label_4 = QtWidgets.QLabel(self.frame_4)
        self.label_4.setGeometry(QtCore.QRect(0, 30, 32, 32))
        self.label_4.setText("")
        self.label_4.setPixmap(QtGui.QPixmap(":/res/res/0_00.png"))
        self.label_4.setObjectName("label_4")
        self.label_5 = QtWidgets.QLabel(self.frame_4)
        self.label_5.setGeometry(QtCore.QRect(84, 30, 32, 32))
        self.label_5.setText("")
        self.label_5.setPixmap(QtGui.QPixmap(":/res/res/4_00.png"))
        self.label_5.setObjectName("label_5")
        self.label_6 = QtWidgets.QLabel(self.frame_4)
        self.label_6.setGeometry(QtCore.QRect(170, 30, 32, 32))
        self.label_6.setText("")
        self.label_6.setPixmap(QtGui.QPixmap(":/res/res/8_00.png"))
        self.label_6.setObjectName("label_6")
        self.label_7 = QtWidgets.QLabel(self.frame_4)
        self.label_7.setGeometry(QtCore.QRect(422, 30, 32, 32))
        self.label_7.setText("")
        self.label_7.setPixmap(QtGui.QPixmap(":/res/res/20_00.png"))
        self.label_7.setObjectName("label_7")
        self.label_8 = QtWidgets.QLabel(self.frame_4)
        self.label_8.setGeometry(QtCore.QRect(510, 30, 32, 32))
        self.label_8.setText("")
        self.label_8.setPixmap(QtGui.QPixmap(":/res/res/0_00.png"))
        self.label_8.setObjectName("label_8")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setGeometry(QtCore.QRect(20, 35, 261, 31))
        self.label.setStyleSheet("font-size: 20px; color: #2aa877;\n" "font-weight: 100;")
        self.label.setObjectName("label")
        self.time_float_label = QtWidgets.QLabel(self.centralwidget)
        self.time_float_label.setGeometry(QtCore.QRect(350, 47, 41, 18))
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setPointSize(-1)
        font.setBold(False)
        font.setWeight(37)
        self.time_float_label.setFont(font)
        self.time_float_label.setStyleSheet("color: #777; font-size: 12px;")
        self.time_float_label.setObjectName("time_float_label")
        self.time_label = QtWidgets.QLabel(self.centralwidget)
        self.time_label.setGeometry(QtCore.QRect(350, 20, 81, 31))
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setPointSize(-1)
        font.setBold(False)
        font.setWeight(37)
        self.time_label.setFont(font)
        self.time_label.setStyleSheet("color: #444; font-size: 22px;")
        self.time_label.setObjectName("time_label")
        self.label_9 = QtWidgets.QLabel(self.centralwidget)
        self.label_9.setGeometry(QtCore.QRect(20, 16, 261, 31))
        self.label_9.setStyleSheet("font-size: 13px;\n" "color: #aaa;\n" "font-weight: 600;")
        self.label_9.setObjectName("label_9")
        self.frame_2.raise_()
        self.frame_current_setting.raise_()
        self.frame_4.raise_()
        self.label.raise_()
        self.time_float_label.raise_()
        self.time_label.raise_()
        self.label_9.raise_()
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Day Time Editor"))
        self.lbl_current_setting.setText(_translate("MainWindow", "Occlusion Strength"))
        self.btn_reset.setText(_translate("MainWindow", "Reset curve"))
        self.lbl_setting_desc.setText(
            _translate(
                "MainWindow",
                "Description: Some Description about the setting, to give the user a rough idea what this does. This description may be so long, it can even go into the next line."
            )
        )
        self.btn_insert_point.setText(_translate("MainWindow", "Insert point from data"))
        self.settings_tree.setSortingEnabled(False)
        self.settings_tree.headerItem().setText(0, _translate("MainWindow", "Setting"))
        self.settings_tree.headerItem().setText(1, _translate("MainWindow", "Value"))
        __sortingEnabled = self.settings_tree.isSortingEnabled()
        self.settings_tree.setSortingEnabled(False)
        self.settings_tree.topLevelItem(0).setText(0, _translate("MainWindow", "Scattering"))
        self.settings_tree.topLevelItem(0).child(0).setText(0, _translate("MainWindow", "Sun Height"))
        self.settings_tree.topLevelItem(0).child(0).setText(1, _translate("MainWindow", "0,4"))
        self.settings_tree.topLevelItem(0).child(1).setText(0, _translate("MainWindow", "Sun Color"))
        self.settings_tree.topLevelItem(0).child(1).setText(1, _translate("MainWindow", "[128, 255, 100]"))
        self.settings_tree.topLevelItem(0).child(2).setText(0, _translate("MainWindow", "Sun Angle"))
        self.settings_tree.topLevelItem(0).child(2).setText(1, _translate("MainWindow", "180 °"))
        self.settings_tree.topLevelItem(1).setText(0, _translate("MainWindow", "Ambient Occlusion"))
        self.settings_tree.topLevelItem(1).child(0).setText(0, _translate("MainWindow", "Occlusion Strength"))
        self.settings_tree.topLevelItem(1).child(0).setText(1, _translate("MainWindow", "0.5"))
        self.settings_tree.setSortingEnabled(__sortingEnabled)
        self.label.setText(_translate("MainWindow", "TIME OF DAY EDITOR"))
        self.time_float_label.setText(_translate("MainWindow", "0.486"))
        self.time_label.setText(_translate("MainWindow", "11:15"))
        self.label_9.setText(_translate("MainWindow", "RENDER PIPELINE"))


from . import resources_rc



from metadrive.engine.logger import get_logger
from direct.filter.FilterManager import FilterManager
import panda3d.core as p3d
from simplepbr import _load_shader_str
from typing import Union, List

import numpy as np
from panda3d.core import NodePath, Vec3, Vec4, Camera, PNMImage, Shader, RenderState, ShaderAttrib

from metadrive.constants import RENDER_MODE_ONSCREEN, BKG_COLOR, RENDER_MODE_NONE


class ImageBuffer:
    LINE_FRAME_COLOR = (0.8, 0.8, 0.8, 0)
    CAM_MASK = None
    BUFFER_W = 84  # left to right
    BUFFER_H = 84  # bottom to top
    BKG_COLOR = BKG_COLOR
    # display_bottom = 0.8
    # display_top = 1
    display_region = None
    display_region_size = [1 / 3, 2 / 3, 0.8, 1.0]
    line_borders = []

    def __init__(
        self,
        width: float,
        height: float,
        pos: Vec3,
        bkg_color: Union[Vec4, Vec3],
        parent_node: NodePath = None,
        frame_buffer_property=None,
        setup_pbr=False,
        engine=None
    ):
        self.logger = get_logger(self.__class__.__name__)
        self._node_path_list = []

        # from metadrive.engine.engine_utils import get_engine
        self.engine = engine
        try:
            assert self.engine.win is not None, "{} cannot be made without use_render or image_observation".format(
                self.__class__.__name__
            )
            assert self.CAM_MASK is not None, "Define a camera mask for every image buffer"
        except AssertionError:
            self.logger.debug("Cannot create {}".format(self.__class__.__name__))
            self.buffer = None
            self.cam = NodePath(Camera("non-sense camera"))
            self._node_path_list.append(self.cam)

            self.lens = self.cam.node().getLens()
            return

        # self.texture = Texture()
        if frame_buffer_property is None:
            self.buffer = self.engine.win.makeTextureBuffer("camera", width, height)
        else:
            self.buffer = self.engine.win.makeTextureBuffer("camera", width, height, fbp=frame_buffer_property)
            # now we have to setup a new scene graph to make this scene

        self.origin = NodePath("new render")

        # this takes care of setting up their camera properly
        self.cam = self.engine.makeCamera(self.buffer, clearColor=bkg_color)
        self.cam.setPos(pos)
        # should put extrinsic parameters here
        self.cam.reparentTo(self.origin)
        # self.cam.setH(-90)  # face to x
        self.lens = self.cam.node().getLens()
        self.cam.node().setCameraMask(self.CAM_MASK)
        if parent_node is not None:
            self.origin.reparentTo(parent_node)
        self.scene_tex = None
        if setup_pbr:
            self.manager = FilterManager(self.buffer, self.cam)
            fbprops = p3d.FrameBufferProperties()
            fbprops.float_color = True
            fbprops.set_rgba_bits(16, 16, 16, 16)
            fbprops.set_depth_bits(24)
            fbprops.set_multisamples(self.engine.pbrpipe.msaa_samples)
            self.scene_tex = p3d.Texture()
            self.scene_tex.set_format(p3d.Texture.F_rgba16)
            self.scene_tex.set_component_type(p3d.Texture.T_float)
            self.tonemap_quad = self.manager.render_scene_into(colortex=self.scene_tex, fbprops=fbprops)
            #
            defines = {}
            #
            post_vert_str = _load_shader_str('post.vert', defines)
            post_frag_str = _load_shader_str('tonemap.frag', defines)
            tonemap_shader = p3d.Shader.make(
                p3d.Shader.SL_GLSL,
                vertex=post_vert_str,
                fragment=post_frag_str,
            )
            self.tonemap_quad.set_shader(tonemap_shader)
            self.tonemap_quad.set_shader_input('tex', self.scene_tex)
            self.tonemap_quad.set_shader_input('exposure', 1.0)

        self.logger.debug("Load Image Buffer: {}".format(self.__class__.__name__))

    def get_image(self):
        """
        Bugs here! when use offscreen mode, thus the front cam obs is not from front cam now
        """
        img = PNMImage()
        self.buffer.getDisplayRegions()[1].getScreenshot(img)
        return img

    def save_image(self, name="debug.png"):
        """
        for debug use
        """
        img = self.get_image()
        img.write(name)

    def get_rgb_array_cpu(self):
        origin_img = self.buffer.getDisplayRegion(1).getScreenshot()
        img = np.frombuffer(origin_img.getRamImage().getData(), dtype=np.uint8)
        img = img.reshape((origin_img.getYSize(), origin_img.getXSize(), 4))
        # img = np.swapaxes(img, 1, 0)
        img = img[::-1]
        img = img[..., :-1]
        return img

    @staticmethod
    def get_grayscale_array(img, clip=True):
        raise DeprecationWarning("This API is deprecated")
        if not clip:
            numpy_array = np.array(
                [[int(img.getGray(i, j) * 255) for j in range(img.getYSize())] for i in range(img.getXSize())],
                dtype=np.uint8
            )
            return np.clip(numpy_array, 0, 255)
        else:
            numpy_array = np.array([[img.getGray(i, j) for j in range(img.getYSize())] for i in range(img.getXSize())])
            return np.clip(numpy_array, 0, 1)

    def add_display_region(self, display_region: List[float]):
        if self.engine.mode != RENDER_MODE_NONE and self.display_region is None:
            # only show them when onscreen
            self.display_region = self.engine.win.makeDisplayRegion(*display_region)
            self.display_region.setCamera(self.buffer.getDisplayRegions()[1].camera)
            self.draw_border(display_region)

    def draw_border(self, display_region):
        engine = self.engine
        # add white frame for display region, convert to [-1, 1]
        left = display_region[0] * 2 - 1
        right = display_region[1] * 2 - 1
        bottom = display_region[2] * 2 - 1
        top = display_region[3] * 2 - 1

        self.line_borders.append(engine.draw_line_2d([left, bottom], [left, top], self.LINE_FRAME_COLOR, 1.5))
        self.line_borders.append(engine.draw_line_2d([left, top], [right, top], self.LINE_FRAME_COLOR, 1.5))
        self.line_borders.append(engine.draw_line_2d([right, top], [right, bottom], self.LINE_FRAME_COLOR, 1.5))
        self.line_borders.append(engine.draw_line_2d([right, bottom], [left, bottom], self.LINE_FRAME_COLOR, 1.5))

    def remove_display_region(self):
        engine = self.engine
        if engine.mode == RENDER_MODE_ONSCREEN and self.display_region is not None:
            engine.win.removeDisplayRegion(self.display_region)
            self.display_region = None
        for line_node in self.line_borders:
            line_node.detachNode()

    def destroy(self):
        engine = self.engine
        if engine is not None:
            self.remove_display_region()
            if self.buffer is not None:
                engine.graphicsEngine.removeWindow(self.buffer)
            self.display_region = None
            self.buffer = None
            if self.cam in engine.camList:
                engine.camList.remove(self.cam)
        self.cam.removeNode()
        if len(self.line_borders) != 0:
            for line_np in self.line_borders:
                if line_np:
                    line_np.removeNode()
        self.line_borders = []
        if hasattr(self, "origin"):
            self.origin.removeNode()

        from metadrive.base_class.base_object import clear_node_list
        clear_node_list(self._node_path_list)

    def __del__(self):
        self.logger.debug("{} is destroyed".format(self.__class__.__name__))

    @classmethod
    def update_display_region_size(cls, display_region_size):
        cls.display_region_size = display_region_size



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six import iteritems

from panda3d.core import Texture, Vec3

from direct.gui.DirectScrolledFrame import DirectScrolledFrame
from direct.gui.DirectFrame import DirectFrame

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.util.generic import rgb_from_string
from metadrive.render_pipeline.rpcore.util.display_shader_builder import DisplayShaderBuilder
from metadrive.render_pipeline.rpcore.util.shader_input_blocks import SimpleInputBlock, GroupedInputBlock

from metadrive.render_pipeline.rpcore.gui.draggable_window import DraggableWindow
from metadrive.render_pipeline.rpcore.gui.text import Text
from metadrive.render_pipeline.rpcore.gui.sprite import Sprite


class PipeViewer(DraggableWindow):
    """ Small tool which displays the order of the graphic pipes """

    _STAGE_MGR = None

    @classmethod
    def register_stage_mgr(cls, mgr):
        """ Sets the stage manager, this is a workaround to prevent
        circular imports, since the pipe viewer is already included
        from the StageManager """
        cls._STAGE_MGR = mgr

    def __init__(self, pipeline, parent):
        """ Constructs the pipe viewer """
        DraggableWindow.__init__(self, width=1300, height=900, parent=parent, title="Pipeline Visualizer")
        self._pipeline = pipeline
        self._scroll_width = 8000
        self._scroll_height = 3000
        self._created = False
        self._create_components()
        self.hide()

    def toggle(self):
        """ Toggles the pipe viewer """
        if self._visible:
            Globals.base.taskMgr.remove("RP_GUI_UpdatePipeViewer")
            self.hide()
        else:
            Globals.base.taskMgr.add(self._update_task, "RP_GUI_UpdatePipeViewer")
            if not self._created:
                self._populate_content()
            self.show()

    def _update_task(self, task=None):
        """ Updates the viewer """
        scroll_value = self._content_frame.horizontalScroll["value"]
        scroll_value *= 2.45
        self._pipe_descriptions.set_x(scroll_value * 2759.0)
        return task.cont

    def _populate_content(self):  # pylint: disable=too-many-branches,too-many-statements
        """ Reads the pipes and stages from the stage manager and renders those
        into the window """
        self._created = True
        self._pipe_node = self._content_node.attach_new_node("pipes")
        self._pipe_node.set_scale(1, 1, -1)
        self._stage_node = self._content_node.attach_new_node("stages")
        current_pipes = []
        pipe_pixel_size = 3
        pipe_height = 100

        # Generate stages
        for offs, stage in enumerate(self._STAGE_MGR.stages):
            node = self._content_node.attach_new_node("stage")
            node.set_pos(220 + offs * 200.0, 0, 20)
            node.set_scale(1, 1, -1)
            DirectFrame(parent=node, frameSize=(10, 150, 0, -3600), frameColor=(0.2, 0.2, 0.2, 1))
            Text(text=str(stage.debug_name.replace("Stage", "")), parent=node, x=20, y=25, size=15)

            for output_pipe, pipe_tex in iteritems(stage.produced_pipes):
                pipe_idx = 0
                r, g, b = rgb_from_string(output_pipe)
                if output_pipe in current_pipes:
                    pipe_idx = current_pipes.index(output_pipe)
                else:
                    current_pipes.append(output_pipe)
                    pipe_idx = len(current_pipes) - 1
                    DirectFrame(
                        parent=node,
                        frameSize=(0, 8000, pipe_pixel_size / 2, -pipe_pixel_size / 2),
                        frameColor=(r, g, b, 1),
                        pos=(10, 1, -95 - pipe_idx * pipe_height)
                    )
                w = 160
                h = Globals.native_resolution.y /\
                    float(Globals.native_resolution.x) * w
                DirectFrame(
                    parent=node,
                    frameSize=(
                        -pipe_pixel_size, w + pipe_pixel_size, h / 2 + pipe_pixel_size, -h / 2 - pipe_pixel_size
                    ),
                    frameColor=(r, g, b, 1),
                    pos=(0, 1, -95 - pipe_idx * pipe_height)
                )

                if isinstance(pipe_tex, (list, tuple)):
                    pipe_tex = pipe_tex[0]

                if isinstance(pipe_tex, (SimpleInputBlock, GroupedInputBlock)):
                    icon_file = "/$$rp/data/gui/icon_ubo.png"
                elif pipe_tex.get_z_size() > 1:
                    icon_file = "/$$rp/data/gui/icon_texture.png"
                elif pipe_tex.get_texture_type() == Texture.TT_buffer_texture:
                    icon_file = "/$$rp/data/gui/icon_buffer_texture.png"
                else:
                    icon_file = None
                    preview = Sprite(
                        image=pipe_tex,
                        parent=node,
                        x=0,
                        y=50 + pipe_idx * pipe_height,
                        w=w,
                        h=h,
                        any_filter=False,
                        transparent=False
                    )

                    preview_shader = DisplayShaderBuilder.build(pipe_tex, int(w), int(h))
                    preview.set_shader(preview_shader)

                    preview.set_shader_inputs(mipmap=0, slice=0, brightness=1, tonemap=False)

                if icon_file:
                    Sprite(
                        image=icon_file,
                        parent=node,
                        x=55,
                        y=65 + pipe_idx * pipe_height,
                        w=48,
                        h=48,
                        near_filter=False,
                        transparent=True
                    )

                    if isinstance(pipe_tex, (SimpleInputBlock, GroupedInputBlock)):
                        tex_desc = "UBO"
                    else:
                        tex_desc = pipe_tex.format_texture_type(pipe_tex.get_texture_type())
                        tex_desc += " - " + pipe_tex.format_format(pipe_tex.get_format()).upper()

                    Text(
                        text=tex_desc,
                        parent=node,
                        x=55 + 48 / 2,
                        y=130 + pipe_idx * pipe_height,
                        color=Vec3(0.2),
                        size=12,
                        align="center"
                    )

            for input_pipe in stage.required_pipes:
                if input_pipe not in current_pipes:
                    self.warn("Pipe not found:", input_pipe)
                    continue
                idx = current_pipes.index(input_pipe)
                r, g, b = rgb_from_string(input_pipe)
                DirectFrame(
                    parent=node,
                    frameSize=(0, 10, 40, -40),
                    frameColor=(r, g, b, 1),
                    pos=(5, 1, -95 - idx * pipe_height)
                )

        self._pipe_descriptions = self._content_node.attach_new_node("PipeDescriptions")
        self._pipe_descriptions.set_scale(1, 1, -1)

        DirectFrame(parent=self._pipe_descriptions, frameSize=(0, 190, 0, -5000), frameColor=(0.1, 0.1, 0.1, 1.0))

        # Generate the pipe descriptions
        for idx, pipe in enumerate(current_pipes):
            r, g, b = rgb_from_string(pipe)
            DirectFrame(
                parent=self._pipe_descriptions,
                frameSize=(0, 180, -95, -135),
                frameColor=(r, g, b, 1.0),
                pos=(0, 1, -idx * pipe_height)
            )
            Text(parent=self._pipe_descriptions, text=pipe, x=42, y=121 + idx * pipe_height, size=15, color=Vec3(0.1))
            Sprite(
                parent=self._pipe_descriptions,
                x=9,
                y=103 + idx * pipe_height,
                image="/$$rp/data/gui/icon_pipe.png",
                transparent=True,
                near_filter=False
            )

    def _create_components(self):
        """ Internal method to create the window components """
        DraggableWindow._create_components(self)

        self._content_frame = DirectScrolledFrame(
            frameSize=(0, self._width - 40, 0, self._height - 80),
            canvasSize=(0, self._scroll_width, 0, self._scroll_height),
            autoHideScrollBars=False,
            scrollBarWidth=20.0,
            frameColor=(0, 0, 0, 0),
            verticalScroll_relief=False,
            horizontalScroll_relief=False,
            parent=self._node,
            pos=(20, 1, -self._height + 20)
        )
        self._content_node = self._content_frame.getCanvas().attach_new_node("PipeComponents")
        self._content_node.set_scale(1, 1, -1)
        self._content_node.set_z(self._scroll_height)



import copy

from typing import Union, Any

import numpy as np

import gym
import gym.spaces
import gymnasium

from metadrive import MetaDriveEnv
from metadrive.envs.gym_wrapper import createGymWrapper, gymToGymnasium, gymnasiumToGym


def test_conversion_functions():
    gym_space_box = gym.spaces.Box(0, 1, shape=(1, 2, 3), dtype=np.float32)
    gymnasium_space_box = gymnasium.spaces.Box(0, 1, shape=(1, 2, 3), dtype=np.float32)
    assert gymToGymnasium(gym_space_box) == gymnasium_space_box
    assert gymnasiumToGym(gymnasium_space_box) == gym_space_box

    gym_space_discrete = gym.spaces.Discrete(5)
    gymnasium_space_discrete = gymnasium.spaces.Discrete(5)
    assert gymToGymnasium(gym_space_discrete) == gymnasium_space_discrete
    assert gymnasiumToGym(gymnasium_space_discrete) == gym_space_discrete

    gym_space_multi_discrete = gym.spaces.MultiDiscrete([5, 6, 7])
    gymnasium_space_multi_discrete = gymnasium.spaces.MultiDiscrete([5, 6, 7])
    assert gymToGymnasium(gym_space_multi_discrete) == gymnasium_space_multi_discrete
    assert gymnasiumToGym(gymnasium_space_multi_discrete) == gym_space_multi_discrete

    gym_space_tuple = gym.spaces.Tuple([gym_space_box, gym_space_discrete])
    gymnasium_space_tuple = gymnasium.spaces.Tuple([gymnasium_space_box, gymnasium_space_discrete])
    assert gymToGymnasium(gym_space_tuple) == gymnasium_space_tuple
    assert gymnasiumToGym(gymnasium_space_tuple) == gym_space_tuple

    gym_space_dict = gym.spaces.Dict({"a": gym_space_box, "b": gym_space_discrete})
    gymnasium_space_dict = gymnasium.spaces.Dict({"a": gymnasium_space_box, "b": gymnasium_space_discrete})
    assert gymToGymnasium(gym_space_dict) == gymnasium_space_dict
    assert gymnasiumToGym(gymnasium_space_dict) == gym_space_dict


# test that mode is removed from args
class NoModeGymnasiumInnerClass(gymnasium.Env):
    def __init__(self, config: dict):
        super().__init__()
        self.action_space = gymnasium.spaces.Discrete(5)
        self.observation_space = gymnasium.spaces.Box(0, 1, shape=(1, 2, 3), dtype=np.float32)

    def render(self, **kwargs):
        assert 'mode' not in kwargs.keys()


def test_gym_wrapper_removes_mode():
    gym_env = createGymWrapper(NoModeGymnasiumInnerClass)(config={})
    gym_env.render(mode=1)


class ConfigOverrideGymnasiumInnerClass(gymnasium.Env):
    @classmethod
    def default_config(cls) -> dict:
        return {"a": 1, "b": 2}

    def __init__(self, config: dict):
        super().__init__()
        assert self.default_config() == {"a": 1, "b": 2}


def test_config_override():
    # should not raise
    createGymWrapper(ConfigOverrideGymnasiumInnerClass)(config={})

    # if we override the default config it should be visible in the (non-overriden) constructor
    try:
        GymWrapperedClass = createGymWrapper(ConfigOverrideGymnasiumInnerClass)

        class OverridenGymWrapperedClass_1(GymWrapperedClass):
            @classmethod
            def default_config(cls) -> dict:
                return {"a": 3, "b": 4}

        # try to initialize (should raise)
        OverridenGymWrapperedClass_1(config={})
    except AssertionError:
        pass
    else:
        raise AssertionError("OverridenGymWrapperedClass_1 should have raised an AssertionError")

    # if we override the default config it should be visible in the overriden constructor
    GymWrapperedClass = createGymWrapper(ConfigOverrideGymnasiumInnerClass)

    class OverridenGymWrapperedClass_2(GymWrapperedClass):
        @classmethod
        def default_config(cls) -> dict:
            return {"a": 3, "b": 4}

        def __init__(self, config: dict):
            assert self.default_config() == {"a": 3, "b": 4}

    # try to initialize (should not raise)
    OverridenGymWrapperedClass_2(config={})


class SingleAgentEnvironment(gymnasium.Env):
    def __init__(self, config: dict):
        super().__init__()
        self.action_space = gymnasium.spaces.Box(-1, 1, shape=(2, ))
        self.observation_space = gymnasium.spaces.Box(0, 1, shape=(1, 2, 3), dtype=np.float32)

        self.active = True
        self.steps = 0
        self.reward_source = gymnasium.spaces.Box(0, 100, dtype=np.float32)

    def step(self, action: np.ndarray):
        assert self.active
        assert self.action_space.contains(action)

        self.steps += 1

        # truncate if steps above 50
        if self.steps < 50:
            # terminate if action is less than 0
            if action[0] < 0:
                self.active = False
                return self.observation_space.sample(), float(self.reward_source.sample().item()), True, False, {}
            else:
                return self.observation_space.sample(), float(self.reward_source.sample().item()), False, False, {}
        else:
            self.active = False
            return self.observation_space.sample(), float(self.reward_source.sample().item()), False, True, {}

    def reset(self, seed: Union[int, None] = None):
        super().reset(seed=seed)
        self.steps = 0
        self.observation_space.seed(seed)
        self.reward_source.seed(seed)
        self.active = True
        return self.observation_space.sample(), {}


def test_gym_wrapper_single_agent():
    # collect data from gymnasium env
    gymnasium_env = SingleAgentEnvironment(config={})
    gym_env = createGymWrapper(SingleAgentEnvironment)(config={})

    assert gymToGymnasium(gym_env.observation_space) == gymnasium_env.observation_space
    assert gymToGymnasium(gym_env.action_space) == gymnasium_env.action_space

    o_gymnasium, _ = gymnasium_env.reset(seed=0)
    o_gym = gym_env.reset(seed=0)

    assert isinstance(o_gymnasium, np.ndarray)
    assert isinstance(o_gym, np.ndarray)

    assert np.allclose(o_gymnasium, o_gym)

    # test will end due to truncation
    while True:
        o_gymnasium, r_gymnasium, tm_gymnasium, tc_gymnasium, _ = gymnasium_env.step(
            np.array([0.0, 0.0], dtype=np.float32)
        )
        o_gym, r_gym, d_gym, _ = gym_env.step(np.array([0.0, 0.0], dtype=np.float32))

        assert isinstance(o_gymnasium, np.ndarray)
        assert isinstance(o_gym, np.ndarray)

        assert np.allclose(o_gymnasium, o_gym)

        assert isinstance(r_gymnasium, float)
        assert isinstance(r_gym, float)

        assert np.allclose(r_gymnasium, r_gym)

        assert d_gym == tm_gymnasium or tc_gymnasium

        if d_gym:
            break

    # now test termination
    o_gymnasium, _ = gymnasium_env.reset(seed=42)
    o_gym = gym_env.reset(seed=42)

    assert isinstance(o_gymnasium, np.ndarray)
    assert isinstance(o_gym, np.ndarray)

    assert np.allclose(o_gymnasium, o_gym)

    # this should terminate
    o_gymnasium, r_gymnasium, tm_gymnasium, tc_gymnasium, _ = gymnasium_env.step(np.array([-0.5, 0], dtype=np.float32))
    o_gym, r_gym, d_gym, _ = gym_env.step(np.array([-0.5, 0], dtype=np.float32))

    assert isinstance(o_gymnasium, np.ndarray)
    assert isinstance(o_gym, np.ndarray)
    assert np.allclose(o_gymnasium, o_gym)

    assert isinstance(r_gymnasium, float)
    assert isinstance(r_gym, float)
    assert np.allclose(r_gymnasium, r_gym)

    assert not tc_gymnasium
    assert tm_gymnasium

    assert d_gym


class MultiAgentEnvironment(gymnasium.Env):
    def __init__(self, config: dict):
        super().__init__()

        # validate config
        assert "num_agents" in config
        assert isinstance(config["num_agents"], int)
        assert "horizon" in config
        assert isinstance(config["horizon"], int)

        self.num_agents = config["num_agents"]
        self.horizon = config["horizon"]

        self.action_space = gymnasium.spaces.Box(-1, 1, shape=(2, ))
        self.observation_space = gymnasium.spaces.Box(0, 1, shape=(1, 2, 3), dtype=np.float32)

        self.steps = 0
        self.reward_source = gymnasium.spaces.Box(0, 100, dtype=np.float32)

    def _generate_observations(self) -> dict[str, np.ndarray]:
        return {agent: self.observation_space.sample() for agent in sorted(self.agents)}

    def _generate_rewards(self) -> dict[str, float]:
        return {agent: float(self.reward_source.sample().item()) for agent in sorted(self.agents)}

    def _validate_actions(self, actions: dict[str, np.ndarray]) -> bool:
        if set(actions.keys()) != self.agents:
            return False
        return all(self.action_space.contains(action) for action in actions.values())

    def step(self, action: dict[str, np.ndarray]):
        assert len(self.agents) > 0
        assert self._validate_actions(action)

        obss = self._generate_observations()
        rews = self._generate_rewards()
        # terminate if action is less than 0
        terminateds = {agent: action[agent][0] < 0 for agent in self.agents}
        # truncate if steps above horizon
        truncateds = {agent: self.steps >= self.horizon and not terminateds[agent] for agent in self.agents}
        infos = {agent: {} for agent in self.agents}

        # increment steps
        self.steps += 1
        # remove dead agents
        self.agents = {agent for agent in self.agents if not terminateds[agent] and not truncateds[agent]}

        # add __all__ key to terminateds and truncateds
        terminateds["__all__"] = all(terminateds.values())
        truncateds["__all__"] = all(truncateds.values())

        return obss, rews, terminateds, truncateds, infos

    def reset(self, seed: Union[int, None] = None):
        super().reset(seed=seed)
        self.steps = 0
        self.agents = {f"agent_{i}" for i in range(self.num_agents)}
        self.observation_space.seed(seed)
        self.reward_source.seed(seed)
        return self._generate_observations(), {agent: {} for agent in self.agents}


def assert_allclose_dict(a: dict[str, np.ndarray], b: dict[str, np.ndarray]):
    a_keys = sorted(a.keys())
    b_keys = sorted(b.keys())

    assert a_keys == b_keys

    for key in a_keys:
        assert np.allclose(a[key], b[key])


def test_gym_wrapper_multi_agent():
    # test multi agent metadrive environment
    gymnasium_env = MultiAgentEnvironment(config={"num_agents": 5, "horizon": 50})
    gym_env = createGymWrapper(MultiAgentEnvironment)(config={"num_agents": 5, "horizon": 50})

    assert gymToGymnasium(gym_env.observation_space) == gymnasium_env.observation_space
    assert gymToGymnasium(gym_env.action_space) == gymnasium_env.action_space

    o_gymnasium, _ = gymnasium_env.reset(seed=0)
    o_gym = gym_env.reset(seed=0)

    assert isinstance(o_gymnasium, dict)
    assert isinstance(o_gym, dict)
    assert_allclose_dict(o_gymnasium, o_gym)

    # test will end due to truncation
    while True:
        # action dict: generate [0, 0] for each agent
        a_dict = {agent: np.array([0.0, 0.0], dtype=np.float32) for agent in gymnasium_env.agents}

        o_gymnasium, r_gymnasium, tm_gymnasium, tc_gymnasium, _ = gymnasium_env.step(a_dict)
        o_gym, r_gym, d_gym, _ = gym_env.step(a_dict)

        assert isinstance(o_gymnasium, dict)
        assert isinstance(o_gym, dict)
        assert_allclose_dict(o_gymnasium, o_gym)

        assert isinstance(r_gymnasium, dict)
        assert isinstance(r_gym, dict)
        assert_allclose_dict(r_gymnasium, r_gym)

        # show correct termination behavior:
        assert isinstance(tm_gymnasium, dict)
        assert isinstance(tc_gymnasium, dict)
        agentnames = list(set(tm_gymnasium) | set(tc_gymnasium))

        for agent in agentnames:
            assert d_gym[agent] == tm_gymnasium[agent] or tc_gymnasium[agent]

        if all(d_gym.values()):
            break

    # now test termination
    o_gymnasium, _ = gymnasium_env.reset(seed=42)
    o_gym = gym_env.reset(seed=42)

    assert isinstance(o_gymnasium, dict)
    assert isinstance(o_gym, dict)
    assert_allclose_dict(o_gymnasium, o_gym)

    # test will end due to termination of the last agent
    while True:
        # action dict: generate [0, 0] for each agent except the last one (which is [-0.5, 0], so that we can test termination)
        a_dict = {agent: np.array([0.0, 0.0], dtype=np.float32) for agent in gymnasium_env.agents}
        a_dict[sorted(a_dict.keys())[-1]] = np.array([-0.5, 0], dtype=np.float32)

        o_gymnasium, r_gymnasium, tm_gymnasium, tc_gymnasium, _ = gymnasium_env.step(a_dict)
        o_gym, r_gym, d_gym, _ = gym_env.step(a_dict)

        assert isinstance(o_gymnasium, dict)
        assert isinstance(o_gym, dict)
        assert_allclose_dict(o_gymnasium, o_gym)

        assert isinstance(r_gymnasium, dict)
        assert isinstance(r_gym, dict)
        assert_allclose_dict(r_gymnasium, r_gym)

        # show correct termination behavior:
        assert isinstance(tm_gymnasium, dict)
        assert isinstance(tc_gymnasium, dict)
        agentnames = list(set(tm_gymnasium) | set(tc_gymnasium))

        for agent in agentnames:
            assert d_gym[agent] == tm_gymnasium[agent] or tc_gymnasium[agent]

        if all(d_gym.values()):
            break


if __name__ == '__main__':
    test_conversion_functions()
    test_gym_wrapper_removes_mode()
    test_config_override()
    test_gym_wrapper_single_agent()
    test_gym_wrapper_multi_agent()



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from panda3d.core import PNMImage
from panda3d.egg import EggTexture


class TextureMixin(object):
    def get_num_channels(self, filepath):
        image = PNMImage()
        image.read(filepath)
        return image.get_num_channels()

    def get_images(self, material, shader):
        if self._render_type == 'rp':  # custom texture order for RP
            return (
                ((0, 'Diffuse'), self.get_diffuse(material, shader)),
                ((1, 'Normal'), self.get_normal_map(material, shader)),
                ((2, 'Specular'), self.get_specular_map(material, shader)),
                ((3, 'Roughness'), self.get_roughness_map(material, shader)),
            )
        else:
            return (
                ((0, 'Diffuse'), self.get_diffuse(material, shader)),
                ((1, 'Normal'), self.get_normal_map(material, shader)),
            )

    def make_texture(self, type_, image_texture):
        filepath = image_texture.image.filepath.lstrip('/')
        path = os.path.dirname(self._output)

        egg_texture = EggTexture(image_texture.image.name, os.path.join(path, filepath))

        if image_texture.extension == 'CLIP':
            egg_texture.set_wrap_mode(EggTexture.WM_clamp)
        elif image_texture.extension == 'REPEAT':
            egg_texture.set_wrap_mode(EggTexture.WM_repeat)

        # check if we have patched panda3d
        have_srgb = hasattr(EggTexture, 'F_srgb_alpha')
        num_channels = self.get_num_channels(os.path.join(path, filepath))

        if type_[1] == 'Diffuse':
            if have_srgb:
                if num_channels == 4:
                    egg_texture.set_format(EggTexture.F_srgb_alpha)
                elif num_channels == 3:
                    egg_texture.set_format(EggTexture.F_srgb)
            else:
                if num_channels == 4:
                    egg_texture.set_format(EggTexture.F_rgba)
                elif num_channels == 3:
                    egg_texture.set_format(EggTexture.F_rgb)

        else:
            if num_channels == 4:
                egg_texture.set_format(EggTexture.F_rgba)
            elif num_channels == 3:
                egg_texture.set_format(EggTexture.F_rgb)

        if type_[1] == 'Diffuse':
            egg_texture.set_env_type(EggTexture.ET_modulate)
        elif type_[1] == 'Normal':
            egg_texture.set_env_type(EggTexture.ET_normal)
        else:
            egg_texture.set_env_type(EggTexture.ET_decal)

        egg_texture.set_priority(type_[0])
        egg_texture.set_stage_name(type_[1])

        return None, egg_texture  # sampler, image

    def make_empty_texture(self, type_):
        filepath = 'textures/{}.png'.format(type_.lower())
        path = os.path.dirname(self._output)

        egg_texture = EggTexture(os.path.basename(filepath), os.path.join(path, filepath))

        egg_texture.set_wrap_mode(EggTexture.WM_clamp)

        # check if we have patched panda3d
        have_srgb = hasattr(EggTexture, 'F_srgb_alpha')

        if type_[1] == 'Diffuse':
            if have_srgb:
                egg_texture.set_format(EggTexture.F_srgb_alpha)
            else:
                egg_texture.set_format(EggTexture.F_rgba)
        else:
            egg_texture.set_format(EggTexture.F_rgba)

        if type_[1] == 'Diffuse':
            egg_texture.set_env_type(EggTexture.ET_modulate)
        elif type_[1] == 'Normal':
            egg_texture.set_env_type(EggTexture.ET_normal)
        else:
            egg_texture.set_env_type(EggTexture.ET_decal)

        egg_texture.set_priority(type_[0])
        egg_texture.set_stage_name(type_[1])

        return None, egg_texture  # sampler, image



"""
Note (Zhenghao, Dec 27, 2020)

This is a evaluation scripts to run RLLib agent in MetaDrive.
Please install ray==1.0.0 first and prepare your trained agent.
This script is put here for reference only.
Future replacement for the expert might use this script.
"""
import logging
import os
import time
from typing import Dict

import matplotlib
import numpy as np
import ray
from ray.rllib.agents.callbacks import DefaultCallbacks
from ray.rllib.agents.ppo import PPOTrainer
from ray.rllib.env import BaseEnv
from ray.rllib.evaluation import MultiAgentEpisode, RolloutWorker
from ray.rllib.execution.rollout_ops import ParallelRollouts
from ray.rllib.policy import Policy

from metadrive import GeneralizationRacing
from metadrive.constants import TerminationState

matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42


class DrivingCallbacks(DefaultCallbacks):
    def on_episode_start(
        self, *, worker: RolloutWorker, base_env: BaseEnv, policies: Dict[str, Policy], episode: MultiAgentEpisode,
        env_index: int, **kwargs
    ):
        episode.user_data["velocity"] = []
        episode.user_data["steering"] = []
        episode.user_data["step_reward"] = []
        episode.user_data["acceleration"] = []

    def on_episode_step(
        self, *, worker: RolloutWorker, base_env: BaseEnv, episode: MultiAgentEpisode, env_index: int, **kwargs
    ):
        info = episode.last_info_for()
        if info is not None:
            episode.user_data["velocity"].append(info["velocity"])
            episode.user_data["steering"].append(info["steering"])
            episode.user_data["step_reward"].append(info["step_reward"])
            episode.user_data["acceleration"].append(info["acceleration"])

    def on_episode_end(
        self, worker: RolloutWorker, base_env: BaseEnv, policies: Dict[str, Policy], episode: MultiAgentEpisode,
        **kwargs
    ):
        arrive_dest = episode.last_info_for()[TerminationState.SUCCESS]
        crash_vehicle = episode.last_info_for()[TerminationState.CRASH_VEHICLE]
        out_of_road = episode.last_info_for()[TerminationState.OUT_OF_ROAD]
        max_step_rate = not (arrive_dest or crash_vehicle or out_of_road)
        episode.custom_metrics["success_rate"] = float(arrive_dest)
        episode.custom_metrics["crash_vehicle_rate"] = float(crash_vehicle)
        episode.custom_metrics["out_of_road_rate"] = float(out_of_road)
        episode.custom_metrics["max_step_rate"] = float(max_step_rate)
        episode.custom_metrics["velocity_max"] = float(np.max(episode.user_data["velocity"]))
        episode.custom_metrics["velocity_mean"] = float(np.mean(episode.user_data["velocity"]))
        episode.custom_metrics["velocity_min"] = float(np.min(episode.user_data["velocity"]))
        episode.custom_metrics["steering_max"] = float(np.max(episode.user_data["steering"]))
        episode.custom_metrics["steering_mean"] = float(np.mean(episode.user_data["steering"]))
        episode.custom_metrics["steering_min"] = float(np.min(episode.user_data["steering"]))
        episode.custom_metrics["acceleration_min"] = float(np.min(episode.user_data["acceleration"]))
        episode.custom_metrics["acceleration_mean"] = float(np.mean(episode.user_data["acceleration"]))
        episode.custom_metrics["acceleration_max"] = float(np.max(episode.user_data["acceleration"]))
        episode.custom_metrics["step_reward_max"] = float(np.max(episode.user_data["step_reward"]))
        episode.custom_metrics["step_reward_mean"] = float(np.mean(episode.user_data["step_reward"]))
        episode.custom_metrics["step_reward_min"] = float(np.min(episode.user_data["step_reward"]))

    def on_train_result(self, *, trainer, result: dict, **kwargs):
        result["success"] = np.nan
        result[TerminationState.CRASH_VEHICLE] = np.nan
        result["out"] = np.nan
        result[TerminationState.MAX_STEP] = np.nan
        result["length"] = result["episode_len_mean"]
        if "success_rate_mean" in result["custom_metrics"]:
            result["success"] = result["custom_metrics"]["success_rate_mean"]
            result[TerminationState.CRASH_VEHICLE] = result["custom_metrics"]["crash_vehicle_rate_mean"]
            result["out"] = result["custom_metrics"]["out_of_road_rate_mean"]
            result[TerminationState.MAX_STEP] = result["custom_metrics"]["max_step_rate_mean"]


def initialize_ray(local_mode=False, num_gpus=None, test_mode=False, **kwargs):
    os.environ['OMP_NUM_THREADS'] = '1'

    if ray.__version__.split(".")[0] == "1":  # 1.0 version Ray
        if "redis_password" in kwargs:
            redis_password = kwargs.pop("redis_password")
            kwargs["_redis_password"] = redis_password

    ray.init(
        logging_level=logging.ERROR if not test_mode else logging.DEBUG,
        log_to_driver=test_mode,
        local_mode=local_mode,
        num_gpus=num_gpus,
        ignore_reinit_error=True,
        **kwargs
    )
    print("Successfully initialize Ray!")
    try:
        print("Available resources: ", ray.available_resources())
    except Exception:
        pass


def get_trainer(checkpoint_path=None, extra_config=None, num_workers=10):
    config = dict(
        num_gpus=0,
        num_workers=num_workers,
        num_cpus_per_worker=1,
        horizon=1000,
        lr=0.0,
        batch_mode="complete_episodes",
        callbacks=DrivingCallbacks,
        # explore=False,  # Add this line to only use mean for action.

        # Setup the correct environment
        env=GeneralizationRacing,
        env_config=dict(num_scenarios=10000)
    )
    if extra_config:
        config.update(extra_config)
    trainer = PPOTrainer(config=config)
    if checkpoint_path is not None:
        trainer.restore(os.path.expanduser(checkpoint_path))
    return trainer


def evaluate(trainer, num_episodes=20):
    ret_reward = []
    ret_length = []
    ret_success_rate = []
    ret_out_rate = []
    ret_crash_vehicle_rate = []
    start = time.time()
    episode_count = 0
    while episode_count < num_episodes:
        rollouts = ParallelRollouts(trainer.workers, mode="bulk_sync")
        batch = next(rollouts)
        episodes = batch.split_by_episode()

        ret_reward.extend([e["rewards"].sum() for e in episodes])
        ret_length.extend([e.count for e in episodes])
        ret_success_rate.extend([e["infos"][-1][TerminationState.SUCCESS] for e in episodes])
        ret_out_rate.extend([e["infos"][-1][TerminationState.OUT_OF_ROAD] for e in episodes])
        ret_crash_vehicle_rate.extend([e["infos"][-1][TerminationState.CRASH_VEHICLE] for e in episodes])

        episode_count += len(episodes)
        print("Finish {} episodes".format(episode_count))

    ret = dict(
        reward=np.mean(ret_reward),
        length=np.mean(ret_length),
        success_rate=np.mean(ret_success_rate),
        out_rate=np.mean(ret_out_rate),
        crash_vehicle_rate=np.mean(ret_crash_vehicle_rate),
        episode_count=episode_count,
        time=time.time() - start,
    )
    print(
        "We collected {} episodes. Spent: {:.3f} s.\nResult: {}".format(
            episode_count,
            time.time() - start, {k: round(v, 3)
                                  for k, v in ret.items()}
        )
    )
    return ret


if __name__ == '__main__':
    ckpt = "checkpoint_417/checkpoint-417"
    num_episodes = 500

    initialize_ray(test_mode=False)
    trainer = get_trainer(ckpt, num_workers=18)
    ret = evaluate(trainer, num_episodes)
    print("Evaluation result: {}".format(ret))



from metadrive.engine.engine_utils import get_engine, initialize_engine, engine_initialized, close_engine, \
    get_global_config, set_global_random_seed

from metadrive.engine.logger import get_logger
